/*
 * Copyright 2003, 2004 Porchdog Software. All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification,
 *	are permitted provided that the following conditions are met:
 *
 *		1. Redistributions of source code must retain the above copyright notice,
 *		   this list of conditions and the following disclaimer.   
 *		2. Redistributions in binary form must reproduce the above copyright notice,
 *		   this list of conditions and the following disclaimer in the documentation
 *		   and/or other materials provided with the distribution.
 *
 *	THIS SOFTWARE IS PROVIDED BY PORCHDOG SOFTWARE ``AS IS'' AND ANY
 *	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *	IN NO EVENT SHALL THE HOWL PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 *	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 *	OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	The views and conclusions contained in the software and documentation are those
 *	of the authors and should not be interpreted as representing official policies,
 *	either expressed or implied, of Porchdog Software.
 */

#include "linux_autoip.h"
#include <netinet/in.h>
#include <linux/sockios.h>
#include <linux/if_ether.h>
#include <salt/socket.h>
#include <salt/address.h>
#include <salt/vstring.h>
#include <salt/vstdlib.h>
#include <salt/verrno.h>
#include <salt/assert.h>
#include <salt/log.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>


static sw_string g_default_interface_name = "eth0";


static sw_result
sw_linux_autoip_network_interface_socket_event_handler(
         								sw_socket_handler handler,
         								sw_salt           salt,
         								sw_socket         socket,
         								sw_socket_event   events,
         								sw_opaque         extra);



sw_string
sw_platform_autoip_network_interface_default_interface_name()
{
	return g_default_interface_name;
}


sw_result
sw_platform_autoip_network_interface_new(
					sw_autoip_network_interface	*	anif,
					sw_salt									salt,
					sw_network_interface					nif)
{
	sw_linux_autoip_network_interface	self;
	int											res;
	int 											macsock;
	struct ifreq								ifr;
	char 											tname[18];

	if (sw_malloc(self, sw_linux_autoip_network_interface, sizeof(struct _sw_linux_autoip_network_interface)) != SW_OKAY)
	{
		sw_log(SW_LOG_ERROR, "sw_platform_autoip_network_interface_new", "malloc failed\n");
		return SW_E_FAIL;
	}

	if (sw_autoip_network_interface_init(&self->m_super, salt, nif) != SW_OKAY)
	{
		sw_log(SW_LOG_ERROR, "sw_platform_autoip_network_interface_new", "sw_autoip_network_interface_init failed\n");
		return SW_E_FAIL;
	}
	
	/*
	 * create out socket
	 */
	if ((self->m_osock = socket(AF_INET, SOCK_PACKET, htons(ETH_P_ARP))) < 0)
	{
		sw_log(SW_LOG_ERROR, "sw_platform_autoip_network_interface_new()", "unable to create sending arp socket\n");
		return SW_E_INIT;
	}

	/*
	 * create in socket
	 */
	if ((self->m_isock = socket(AF_INET, SOCK_PACKET, htons(ETH_P_ARP))) < 0)
	{
		sw_log(SW_LOG_ERROR, "sw_platform_autoip_network_interface_new()", "unable to create receiving arp socket\n");
		return SW_E_INIT;
	}

	/*
	 * set non-blocking socket
	 */
	res = fcntl(self->m_isock, F_GETFL);
	if ((fcntl(self->m_isock, F_SETFL, res | O_NONBLOCK)) < 0)
	{
		sw_log(SW_LOG_ERROR, "sw_platform_autoip_network_interface_new()", "fcntl failed\n");
		return SW_E_INIT;
	}

	if (sw_tcp_socket_init_with_fd(&self->m_socket, self->m_isock) != SW_OKAY)
	{
		sw_log(SW_LOG_ERROR, "sw_platform_autoip_network_interface_new", "sw_tcp_socket_init_with_fd failed\n");
		return SW_E_INIT;
	}

	SW_TRY(sw_autoip_network_interface_handle_wakeup(&self->m_super));

	*anif = &self->m_super;

	return SW_OKAY;
}


sw_result
sw_platform_autoip_network_interface_delete(
					sw_autoip_network_interface anif)
{
	sw_linux_autoip_network_interface self = (sw_linux_autoip_network_interface) anif;

	sw_autoip_network_interface_fina(&self->m_super);

	sw_platform_autoip_network_interface_stop_monitoring(anif);

	/*
	 * free
	 */
	close(self->m_isock);
	close(self->m_osock);

	/* 
	 * and we're good
	 */
	sw_free(self);

	return SW_OKAY;
}


sw_result
sw_platform_autoip_network_interface_set_mac_address(
		const char * s)
{
	return SW_E_INIT;
}


#if 0

sw_result 
aipif_stop_ip(
		aip_if 	self)
{
	int r;
	int sock_fd;
	struct ifreq ifr;
	char	name[IFNAMSIZ];

	/* create socket */
	if ((sock_fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_IP)) < 0)
	{
		sw_log(SW_LOG_ERROR, "aipif_shutdown()", "unable to create socket\n");
		goto done;
	}

	sw_network_interface_name(self->m_nif, name, IFNAMSIZ);
	strncpy(ifr.ifr_name, name, IFNAMSIZ - 1);
	ifr.ifr_name[IFNAMSIZ - 1] = (char)NULL;

	/* get flags */
	if ((r = ioctl(sock_fd, SIOCGIFFLAGS, &ifr)) < 0)
	{
		sw_log(SW_LOG_ERROR, "aipif_shutdown()", "unable to get flags\n");
		goto done;
	}

	/* set flags to bring interface down */
	ifr.ifr_flags &= ~(/* IFF_UP | */ IFF_BROADCAST | IFF_MULTICAST | IFF_PROMISC);
	if ((r = ioctl(sock_fd, SIOCSIFFLAGS, &ifr)) < 0)
	{
		sw_log(SW_LOG_ERROR, "aipif_shutdown()", "unable to bring interface down\n");
		goto done;
	}

done: 

	close(sock_fd);
}

sw_result 
aipif_start_ip(
		aip_if 				self,
		sw_ipv4_address	ipAddr)
{
	char name[32];

	sw_network_interface_name(self->m_nif, name, 32);

	return sw_network_interface_up(name);
}


sw_result
aipif_get_ip_address(
				aip_if		self,
				sw_saddr	* 	s)
{
	sw_result			result;
	sw_ipv4_address	ip;

	if ((result = sw_network_interface_ipv4_address(self->m_nif, &ip)) == SW_OKAY)
	{
		*s = sw_ipv4_address_saddr(ip);
	}

	return result;
}


sw_result
aipif_get_network_interface(
				aip_if						self,
				sw_network_interface * 	netif)
{
	*netif = self->m_nif;

	return SW_OKAY;
}

sw_result
aipif_set_network_interface(
				aip_if						this,
				sw_network_interface  	netif)
{
	sw_network_interface_fina(this->m_nif);
	this->m_nif = netif;

	return SW_OKAY;
}

sw_ulong
aipif_socket_fd(aip_if self)
{
	return self->m_isock;
}
#endif


sw_result
sw_platform_autoip_network_interface_start_monitoring(
											sw_autoip_network_interface	anif)
{
	sw_linux_autoip_network_interface self = (sw_linux_autoip_network_interface) anif;

	return sw_salt_register_socket(self->m_super.m_salt, self->m_socket, SW_SOCKET_READ, (sw_socket_handler) self, sw_linux_autoip_network_interface_socket_event_handler, (sw_opaque) NULL);
}


sw_result
sw_platform_autoip_network_interface_stop_monitoring(
											sw_autoip_network_interface	anif)
{
	sw_linux_autoip_network_interface self = (sw_linux_autoip_network_interface) anif;

	return sw_salt_unregister_socket(self->m_super.m_salt, self->m_socket);
}

											

sw_result
sw_platform_autoip_network_interface_send_arp_packet(
											sw_autoip_network_interface	anif, 
											sw_arp_packet					*	packet)
{
	sw_linux_autoip_network_interface	self = (sw_linux_autoip_network_interface) anif;
	struct sockaddr 							sndarp;
	char											name[IFNAMSIZ];
	sw_mac_address								mymac;

	/*
	 * get interface name
	 */
	sw_network_interface_name(self->m_super.m_nif, name, IFNAMSIZ);
	strcpy(sndarp.sa_data, name);

	/*
	 * send packet
	 */
	if ((sendto(self->m_osock, packet, sizeof(struct _sw_arp_packet), 0, &sndarp, sizeof(sndarp))) < 0)
	{
		sw_log(SW_LOG_ERROR, "aipif_send_arp_request()", "sendto failed\n");
		return SW_E_INIT;
	}

	return SW_OKAY;
}


sw_result
sw_platform_autoip_network_interface_set_ip_address(
											sw_autoip_network_interface	anif)
{
	sw_linux_autoip_network_interface	self = (sw_linux_autoip_network_interface) anif;
   sw_result      result = SW_OKAY;
   int            r;
   int            sock_fd;
   struct ifreq   ifr;
   struct in_addr saddr;
   char           name[IFNAMSIZ];

   /*
	 * socket for ioctl
	 */
   if ((sock_fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_IP)) < 0)
   {
      sw_log(SW_LOG_ERROR, "aipif_set_ip_address()", "unable to create socket\n");
      goto done;
   }

	/*
	 * initialize the device name
	 */
   sw_memset(&ifr, 0, sizeof(ifr));
   sw_network_interface_name(self->m_super.m_nif, name, IFNAMSIZ);
   strncpy(ifr.ifr_name, name, IFNAMSIZ - 1);
                                                          
	/*
	 * initialize the ip address
	 */
   saddr.s_addr = sw_ipv4_address_saddr(self->m_super.m_ip_addr);
   ifr.ifr_addr.sa_family = AF_INET;
   ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr = saddr;
   if ((r = ioctl(sock_fd, SIOCSIFADDR, &ifr)) < 0)
   {
		/*
		 * most common error
		 */
      if (errno == -EPERM)
      {
         sw_log(SW_LOG_ERROR, "aipif_set_ip_address()", "don't have permission on interface\n");
         goto done;
      }
      else  /* all others */
      {
         sw_log(SW_LOG_ERROR, "aipif_set_ip_address()", "unable to bring interface up\n");
         goto done;
      }
   }

done:

   close(sock_fd);

	return sw_network_interface_set_ipv4_address(self->m_super.m_nif, self->m_super.m_ip_addr);
}


static sw_result
sw_linux_autoip_network_interface_socket_event_handler(
         								sw_socket_handler handler,
         								sw_salt           salt,
         								sw_socket         socket,
         								sw_socket_event   events,
         								sw_opaque         extra)
{
	static sw_mac_address	ProbeSenderMAC = { 0, 0, 0, 0, 0, 0 };

	sw_linux_autoip_network_interface self = (sw_linux_autoip_network_interface) handler;
	sw_result					result = SW_OKAY;
   sw_arp_packet    			packet; 
	char 							smyip[18], spacketip[18];
	char 							smymac[18], spacketmac[18];
	int			bytes;
	
   sw_assert(self != NULL);

	/*
	 * ignore invocations if not our socket
	 */
   if (self->m_isock != sw_socket_fd(socket))
   {
		return SW_OKAY;
   }

	sw_memset(&packet, 0, sizeof(struct _sw_arp_packet));
	if ((bytes = read(self->m_isock, &packet, sizeof(struct _sw_arp_packet))) < 0)
	{
		sw_log(SW_LOG_ERROR, "aipif_recv_arp_request()", "read failed\n");
		return SW_E_INIT;
	}

	if (bytes >= (sizeof(struct _sw_ethernet_header) + sizeof(struct _sw_arp_header)))
	{
		if (((ntohs(packet.m_arp_header.m_op)) == SW_AUTOIP_ARP_REQUEST) || ((ntohs(packet.m_arp_header.m_op)) == SW_AUTOIP_ARP_REPLY))
		{
			sw_autoip_network_interface_read_arp_packet(&self->m_super, &packet);
		}
	}

	return SW_OKAY;
}
