/*
 * Copyright 2003, 2004 Porchdog Software. All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification,
 *	are permitted provided that the following conditions are met:
 *
 *		1. Redistributions of source code must retain the above copyright notice,
 *		   this list of conditions and the following disclaimer.   
 *		2. Redistributions in binary form must reproduce the above copyright notice,
 *		   this list of conditions and the following disclaimer in the documentation
 *		   and/or other materials provided with the distribution.
 *
 *	THIS SOFTWARE IS PROVIDED BY PORCHDOG SOFTWARE ``AS IS'' AND ANY
 *	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *	IN NO EVENT SHALL THE HOWL PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 *	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 *	OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	The views and conclusions contained in the software and documentation are those
 *	of the authors and should not be interpreted as representing official policies,
 *	either expressed or implied, of Porchdog Software.
 */

#define SW_MDNS_MAJOR 	0
#define SW_MDNS_MINOR 	9
#define SW_MDNS_MICRO 	4


#if !defined(WIN32)
#	include <sys/types.h>
#	include <sys/socket.h>
#	include <sys/time.h>
#	if !defined(__VXWORKS__)
#		include <sys/uio.h>
#	endif
#	include <netinet/in.h>
#	include <unistd.h>
#endif

#include <NotOSX/notosx_mdns.h>
#include <discovery/text_record.h>
#include <salt/salt.h>
#include <corby/orb.h>
#include <corby/channel.h>
#include <corby/message.h>
#include <corby/buffer.h>
#include <salt/vstring.h>
#include <salt/vstdlib.h>
#include <salt/verrno.h>
#include <salt/assert.h>
#include <salt/log.h>
#include <stdio.h>


#define CACHE_SIZE 750


static sw_result
sw_mdns_servant_check_version(
				sw_mdns_servant					self,
				const sw_corby_object			client,
				sw_corby_channel					channel,
				sw_octet								major,
				sw_octet								minor, 
				sw_octet								micro);


static sw_result
sw_mdns_servant_add_node(
				sw_mdns_servant					self,
				sw_mdns_servant_client_node	node);
				
				
static sw_mdns_servant_client_node
sw_mdns_servant_lookup_node(
				sw_mdns_servant		self,
				sw_corby_channel		channel,
				sw_ulong					id);
				

static sw_result
sw_mdns_servant_free_node(
				sw_mdns_servant					self,
				sw_mdns_servant_client_node	node);


static sw_result
sw_mdns_servant_stop_publish_domain_cleanup(
				sw_mdns_servant		self,
				sw_mdns_servant_client_node	node);


static sw_result
sw_mdns_servant_stop_publish_host_cleanup(
				sw_mdns_servant		self,
				sw_mdns_servant_client_node	node);


static sw_result
sw_mdns_servant_stop_publish_cleanup(
				sw_mdns_servant		self,
				sw_mdns_servant_client_node	node);


static sw_result
sw_mdns_servant_stop_browse_domains_cleanup(
				sw_mdns_servant		self,
				sw_mdns_servant_client_node	node);


static sw_result
sw_mdns_servant_stop_browse_services_cleanup(
				sw_mdns_servant		self,
				sw_mdns_servant_client_node	node);


static sw_result
sw_mdns_servant_stop_resolve_cleanup(
				sw_mdns_servant		self,
				sw_mdns_servant_client_node	node);
				

static void
sw_mdns_servant_publish_host_callback(
				void							*	inContext, 
				DNSHostRegistrationRef		inRef, 
				DNSStatus						inStatusCode, 
				void							*	inData);


static sw_result
sw_mdns_servant_publish_callback(
				void								*	inContext,
				DNSRegistrationRef            inRef,
				DNSStatus                  	inStatusCode,
				const DNSRegistrationEvent *  inEvent);


static sw_result
sw_mdns_servant_publish_callback2(
				void								*	inContext,
				DNSRegistrationRef            inRef,
				DNSStatus                  	inStatusCode,
				const DNSRegistrationEvent *  inEvent);


static sw_result
sw_mdns_servant_browse_callback(
				void							*	inContext,
				DNSBrowserRef					inRef,
				DNSStatus						inStatusCode,
				const DNSBrowserEvent	*	inEvent);

			
static sw_result
sw_mdns_servant_resolve_callback(
				void							*	inContext,
				DNSResolverRef					inRef,
				DNSStatus                  inStatusCode,
				const DNSResolverEvent	*	inEvent);


static sw_result
sw_mdns_servant_convert_status_to_result(
				DNSStatus						status);


static sw_result
sw_mdns_servant_dispatcher(
            sw_corby_servant    	servant,
            sw_salt					salt,
            sw_corby_orb			orb,
            sw_corby_channel  	channel,
            sw_corby_message  	message,
            sw_corby_buffer   	buffer,
            sw_const_string   	op,
            sw_ulong          	op_len,
            sw_ulong					request_id,
            sw_octet					endian);


static sw_result
sw_mdns_servant_connection_notifier(
				sw_corby_orb_observer			handler,
				sw_salt								salt,
				sw_corby_orb						orb,
				struct _sw_corby_channel	*	channel,
				sw_opaque							extra);


static sw_result
sw_mdns_servant_cleanup(
				sw_mdns_servant			self,
				sw_opaque					extra);



static sw_result
sw_mdns_servant_publish_domain_reply(
				sw_discovery_publish_domain_handler	handler,
				sw_discovery									discovery,
				sw_discovery_publish_status				status,
				sw_discovery_publish_domain_id			id,
				sw_opaque										extra);


static sw_result
sw_mdns_servant_publish_host_reply(
				sw_discovery_publish_host_handler	handler,
				sw_discovery								discovery,
				sw_discovery_publish_status			status,
				sw_discovery_publish_host_id			id,
				sw_opaque									extra);


static sw_result
sw_mdns_servant_publish_reply(
				sw_discovery_publish_handler	handler,
				sw_discovery						discovery,
				sw_discovery_publish_status	status,
				sw_discovery_publish_id		id,
				sw_opaque							extra);
				

static sw_result
sw_mdns_servant_browse_reply(
				sw_discovery_browse_handler	handler,
				sw_discovery						discovery,
				sw_discovery_browse_id			id,
				sw_discovery_browse_status	status,
				sw_const_string					name,
				sw_const_string					type,
				sw_const_string					domain,
				sw_opaque							extra);

			
static sw_result
sw_mdns_servant_resolve_reply(
				sw_discovery_resolve_handler	handler,
				sw_discovery						discovery,
				sw_discovery_resolve_id		id,
				sw_const_string					name,
				sw_const_string					type,
				sw_const_string					domain,
				sw_ipv4_address					address,
				sw_port								port,
				sw_const_string					text_record_string,
				sw_octets							text_record,
				sw_ulong								text_record_len,
				sw_opaque							extra);


static sw_result
sw_mdns_servant_getline(
				sw_mdns_servant		self,
				FILE					*	f,
				sw_string				line,
				sw_ulong				*	len);


static sw_result
sw_mdns_servant_parse_conf_line(
				sw_mdns_servant		self,
				sw_string				line,
				sw_ulong					len,
				sw_string				name,
				sw_string				type,
				sw_string				domain,
				sw_ushort			*	port,
				sw_text_record		*	text_record);


/*
 * static function declarations 
 */
static sw_ulong g_id;

static sw_ulong
sw_mdns_stub_next_id()
{
	if (++g_id == 0)
	{
		++g_id;
	}

	return g_id;
}


static sw_result
sw_mdns_stub_bind(
							sw_mdns_stub		self);


static sw_result
sw_mdns_stub_dispatcher(
							sw_corby_servant	servant,
							sw_salt				salt,
							sw_corby_orb		orb,
							sw_corby_channel	channel,
							sw_corby_message	message,
							sw_corby_buffer	buffer,
							sw_const_string	op,
							sw_ulong				op_len,
							sw_ulong				request_id,
							sw_octet				endian);


static sw_mdns_stub_pending_op
sw_mdns_stub_lookup(
							sw_mdns_stub		self,
							sw_ulong				id);


/*
 * sw_network_service implementation
 * -------------------------
 * This is the client side to publishing services using DNS-SD.
 *
 */
sw_result
sw_mdns_stub_init(
					sw_mdns_stub	*	self,
					sw_salt				salt,
					sw_discovery		discovery,
					sw_ushort			port)
{
	static sw_corby_orb_config config[] =
	{
		{ "iiop", SW_TAG_INTERNET_IOP, "127.0.0.1", 0 },
		{ NULL }
	};

	sw_char			mDNSURL[128];
	sw_result		result;
	
	/*
	 * try to allocate space
	 */
	SW_TRY_GOTO(sw_malloc((*self), sw_mdns_stub, sizeof(struct _sw_mdns_stub)));

	/*
	 * initialize the easy stuff
	 */
	(*self)->m_salt		=	salt;
	(*self)->m_discovery =	discovery;	
	(*self)->m_bound		=	SW_FALSE;

	/*
	 * create an orb so we can handle replies
	 */
	SW_TRY_GOTO(sw_corby_orb_init(&(*self)->m_orb, (*self)->m_salt, config, NULL, NULL, NULL));

	/*
		register our server
	*/
	SW_TRY_GOTO(sw_corby_orb_register_servant((*self)->m_orb, (*self), (sw_corby_servant_cb) sw_mdns_stub_dispatcher, SERVICE_OID, &(*self)->m_self, NULL));

	/*
		create client for mDNS service
	*/
	sprintf(mDNSURL, "iiop://127.0.0.1:%d/dns-sd", (int) port);

	SW_TRY_GOTO(sw_corby_object_init_from_url(&(*self)->m_service, (*self)->m_orb, mDNSURL, NULL, 0));

	result =	SW_OKAY;

exit:

	if (result != SW_OKAY)
	{
	}

	return result;
}


sw_result
sw_mdns_stub_fina(
							sw_mdns_stub	self)
{
	sw_corby_orb_unregister_servant(self->m_orb, SERVICE_OID);
	sw_corby_object_fina(self->m_service);
	sw_corby_orb_fina(self->m_orb);

	return SW_OKAY;
}


sw_result
sw_mdns_stub_check_version(
							sw_mdns_stub	self)
{
	static sw_string			op 		=	"check_version";
	static sw_ulong			op_len	=	14;
	sw_corby_buffer 			buffer;
	sw_result					result;
	struct timeval				tv;
	fd_set						fds;
	int							res;

	/*
	 * very coarse grain thread safe-ing
	 */
	sw_salt_lock(self->m_salt);

	/*
	 * try and bind to service
	 */
	SW_TRY_GOTO(sw_mdns_stub_bind(self));

	/*
	 * now do garden variety initialization
	 */
	SW_TRY_GOTO(sw_corby_object_start_request(self->m_service, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_object(buffer, self->m_self));
	SW_TRY_GOTO(sw_corby_buffer_put_octet(buffer, SW_MDNS_MAJOR));
	SW_TRY_GOTO(sw_corby_buffer_put_octet(buffer, SW_MDNS_MINOR));
	SW_TRY_GOTO(sw_corby_buffer_put_octet(buffer, SW_MDNS_MICRO));
	SW_TRY_GOTO(sw_corby_object_send(self->m_service, buffer, NULL, NULL, NULL));

	/*
	 * now wait for reply...this is a hack, as the only requests I can
	 * make are oneways, and I need to wait for a reply to this
	 */
	tv.tv_sec	=	5;
	tv.tv_usec	=	0;
	FD_ZERO(&fds);
	FD_SET(sw_mdns_stub_socket(self), &fds);
	res									=	select(sw_mdns_stub_socket(self) + 1, &fds, NULL, NULL, &tv);
	self->m_check_version_result	=	SW_E_FAIL;
	
	if (res == 1)
	{
		sw_mdns_stub_read_socket(self);
	}

	result = self->m_check_version_result;

exit:

	sw_salt_unlock(self->m_salt);
	
	return result;
}


sw_result
sw_mdns_stub_publish_domain(
							sw_mdns_stub									self,
							sw_const_string								name,
							sw_discovery_publish_domain_handler	handler,
							sw_discovery_publish_domain_reply		reply,
							sw_opaque										extra,
							sw_discovery_publish_domain_id		*	id)
{
	static sw_string			op 		=	"publish_domain";
	static sw_ulong			op_len	=	15;
	sw_corby_buffer 			buffer;
	sw_result					result;
	sw_mdns_stub_pending_op	node		=	NULL;

	/*
	 * very coarse grain thread safe-ing
	 */
	sw_salt_lock(self->m_salt);

	/*
	 * try and bind to service
	 */
	SW_TRY_GOTO(sw_mdns_stub_bind(self));

	SW_TRY_GOTO(sw_malloc(node, sw_mdns_stub_pending_op, sizeof(struct _sw_mdns_stub_pending_op)));

	/*
	 * now do garden variety initialization
	 */
	node->m_handler			= 	handler;
	node->m_publish_reply	= 	reply;
	node->m_extra				= 	extra;
	node->m_id					=	sw_mdns_stub_next_id();
	*id							=	(sw_opaque) node->m_id;
	
	SW_TRY_GOTO(sw_corby_object_start_request(self->m_service, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, name));
	SW_TRY_GOTO(sw_corby_buffer_put_object(buffer, self->m_self));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) node->m_id));
	SW_TRY_GOTO(sw_corby_object_send(self->m_service, buffer, NULL, NULL, NULL));

	/*
	 * stash the node
	 */
	node->m_txen			= self->m_pending_ops;
	self->m_pending_ops	= node;

	/*
	 * if we got here then we must be good
	 */ 
	result = SW_OKAY;
	
exit:

	if (result != SW_OKAY)
	{
		if (node != NULL)
		{
			sw_free(node);
		}
	}

	sw_salt_unlock(self->m_salt);
	
	return result;
}


sw_result
sw_mdns_stub_stop_publish_domain(
								sw_mdns_stub							self,
								sw_discovery_publish_domain_id	id)
{
	static sw_string	op			=	"stop_publish_domain";
	static sw_ulong	op_len	=	20;
	sw_corby_buffer	buffer;
	sw_result			result;

	sw_salt_lock(self->m_salt);

	/*
	 * try and bind to service
	 */
	SW_TRY_GOTO(sw_mdns_stub_bind(self));

	
	SW_TRY_GOTO(sw_corby_object_start_request(self->m_service, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) id));
	SW_TRY_GOTO(sw_corby_object_send(self->m_service, buffer, NULL, NULL, NULL));

	result = SW_OKAY;

exit:

	sw_salt_unlock(self->m_salt);
	
	return result;
}


sw_result
sw_mdns_stub_publish_host(
						sw_mdns_stub								self,
						sw_const_string							name,
						sw_const_string							domain,
						sw_ipv4_address							address,
						sw_discovery_publish_host_handler	handler,
						sw_discovery_publish_host_reply		reply,
						sw_opaque									extra,
						sw_discovery_publish_host_id		*	id)
{
	static sw_string				op 		=	"publish_host";
	static sw_ulong				op_len	=	13;
	sw_corby_buffer 				buffer;
	sw_result						result;
	sw_mdns_stub_pending_op	node		=	NULL;

	/*
	 * very coarse grain thread safe-ing
	 */
	sw_salt_lock(self->m_salt);

	/*
	 * try and bind to service
	 */
	SW_TRY_GOTO(sw_mdns_stub_bind(self));

	SW_TRY_GOTO(sw_malloc(node, sw_mdns_stub_pending_op, sizeof(struct _sw_mdns_stub_pending_op)));

	/*
	 * now do garden variety initialization
	 */
	node->m_handler				= 	handler;
	node->m_publish_host_reply	= 	reply;
	node->m_extra					= 	extra;
	node->m_id						=	sw_mdns_stub_next_id();
	*id								=	(sw_opaque) node->m_id;
	
	SW_TRY_GOTO(sw_corby_object_start_request(self->m_service, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, name));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, domain));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, sw_ipv4_address_saddr(address)));
	SW_TRY_GOTO(sw_corby_buffer_put_object(buffer, self->m_self));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) node->m_id));
	SW_TRY_GOTO(sw_corby_object_send(self->m_service, buffer, NULL, NULL, NULL));

	/*
	 * stash the node
	 */
	node->m_txen			= self->m_pending_ops;
	self->m_pending_ops	= node;

	/*
	 * if we got here then we must be good
	 */ 
	result = SW_OKAY;
	
exit:

	if (result != SW_OKAY)
	{
		if (node != NULL)
		{
			sw_free(node);
		}
	}

	sw_salt_unlock(self->m_salt);
	
	return result;
}


sw_result
sw_mdns_stub_stop_publish_host(
								sw_mdns_stub						self,
								sw_discovery_publish_host_id	id)
{
	static sw_string	op			=	"stop_publish_host";
	static sw_ulong	op_len	=	17;
	sw_corby_buffer	buffer;
	sw_result			result;

	sw_salt_lock(self->m_salt);

	/*
	 * try and bind to service
	 */
	SW_TRY_GOTO(sw_mdns_stub_bind(self));

	
	SW_TRY_GOTO(sw_corby_object_start_request(self->m_service, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) id));
	SW_TRY_GOTO(sw_corby_object_send(self->m_service, buffer, NULL, NULL, NULL));

	result = SW_OKAY;

exit:

	sw_salt_unlock(self->m_salt);
	
	return result;
}


sw_result
sw_mdns_stub_publish(
						sw_mdns_stub						self,
						sw_const_string					name,
						sw_const_string					reg_type,
						sw_const_string					domain,
						sw_const_string					host,
						sw_port								port,
						sw_octets							text_record,
						sw_ulong								text_record_len,
						sw_discovery_publish_handler	handler,
						sw_discovery_publish_reply	reply,
						sw_opaque							extra,
						sw_discovery_publish_id	*	id)
{
	static sw_string				op 		=	"publish";
	static sw_ulong				op_len	=	8;
	sw_corby_buffer 				buffer;
	sw_result						result;
	sw_mdns_stub_pending_op	node		=	NULL;

	/*
	 * very coarse grain thread safe-ing
	 */
	sw_salt_lock(self->m_salt);

	/*
	 * try and bind to service
	 */
	SW_TRY_GOTO(sw_mdns_stub_bind(self));

	SW_TRY_GOTO(sw_malloc(node, sw_mdns_stub_pending_op, sizeof(struct _sw_mdns_stub_pending_op)));

	/*
	 * now do garden variety initialization
	 */
	node->m_handler			= 	handler;
	node->m_publish_reply	= 	reply;
	node->m_extra				= 	extra;
	node->m_id					=	sw_mdns_stub_next_id();
	*id							=	(sw_opaque) node->m_id;
	
	SW_TRY_GOTO(sw_corby_object_start_request(self->m_service, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, name));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, reg_type));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, domain));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, host));
	SW_TRY_GOTO(sw_corby_buffer_put_ushort(buffer, port));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, text_record_len));
	SW_TRY_GOTO(sw_corby_buffer_put_buf(buffer, text_record, text_record_len));
	SW_TRY_GOTO(sw_corby_buffer_put_object(buffer, self->m_self));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) node->m_id));
	SW_TRY_GOTO(sw_corby_object_send(self->m_service, buffer, NULL, NULL, NULL));

	/*
	 * stash the node
	 */
	node->m_txen			= self->m_pending_ops;
	self->m_pending_ops	= node;

	/*
	 * if we got here then we must be good
	 */ 
	result = SW_OKAY;
	
exit:

	if (result != SW_OKAY)
	{
		if (node != NULL)
		{
			sw_free(node);
		}
	}

	sw_salt_unlock(self->m_salt);
	
	return result;
}


sw_result
sw_mdns_stub_publish_update(
						sw_mdns_stub						self,
						sw_discovery_publish_id			id,
						sw_octets							text_record,
						sw_ulong								text_record_len)
{
	static sw_string				op 		=	"publish_update";
	static sw_ulong				op_len	=	15;
	sw_corby_buffer 				buffer;
	sw_result						result;
	sw_mdns_stub_pending_op	node		=	NULL;

	/*
	 * very coarse grain thread safe-ing
	 */
	sw_salt_lock(self->m_salt);

	/*
	 * try and bind to service
	 */
	SW_TRY_GOTO(sw_mdns_stub_bind(self));

	SW_TRY_GOTO(sw_corby_object_start_request(self->m_service, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) id));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, text_record_len));
	SW_TRY_GOTO(sw_corby_buffer_put_buf(buffer, text_record, text_record_len));
	SW_TRY_GOTO(sw_corby_object_send(self->m_service, buffer, NULL, NULL, NULL));

	/*
	 * if we got here then we must be good
	 */ 
	result = SW_OKAY;
	
exit:

	sw_salt_unlock(self->m_salt);
	
	return result;
}


sw_result
sw_mdns_stub_stop_publish(
								sw_mdns_stub				self,
								sw_discovery_publish_id	id)
{
	static sw_string	op			=	"stop_publish";
	static sw_ulong	op_len	=	13;
	sw_corby_buffer	buffer;
	sw_result			result;

	sw_salt_lock(self->m_salt);

	/*
	 * try and bind to service
	 */
	SW_TRY_GOTO(sw_mdns_stub_bind(self));

	
	SW_TRY_GOTO(sw_corby_object_start_request(self->m_service, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) id));
	SW_TRY_GOTO(sw_corby_object_send(self->m_service, buffer, NULL, NULL, NULL));

	result = SW_OKAY;

exit:

	sw_salt_unlock(self->m_salt);
	
	return result;
}


sw_result
sw_mdns_stub_browse_domains(
								sw_mdns_stub						self,
								sw_discovery_browse_handler	handler,
								sw_discovery_browse_reply		reply,
								sw_opaque							extra,
								sw_discovery_browse_id		*	id)
{
	static sw_string				op 		=	"browse_domains";
	static sw_ulong				op_len	=	15;
	sw_corby_buffer 				buffer;
	sw_result						result;
	sw_mdns_stub_pending_op	node		=	NULL;

	sw_salt_lock(self->m_salt);
	
	/*
	 * try and bind to service
	 */
	SW_TRY_GOTO(sw_mdns_stub_bind(self));

	/*
	 * try to allocate space
	 */
	SW_TRY_GOTO(sw_malloc(node, sw_mdns_stub_pending_op, sizeof(struct _sw_mdns_stub_pending_op)));

	/*
	 * now do garden variety initialization
	 */
	node->m_handler			=	handler;
	node->m_browse_reply		=	reply;
	node->m_extra				=	extra;
	node->m_id					=	sw_mdns_stub_next_id();
	*id							=	(sw_opaque) node->m_id;

	SW_TRY_GOTO(sw_corby_object_start_request(self->m_service, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_object(buffer, self->m_self));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) node->m_id));
	SW_TRY_GOTO(sw_corby_object_send(self->m_service, buffer, NULL, NULL, NULL));
	
	/*
	 * stash the node
	 */
	node->m_txen			= self->m_pending_ops;
	self->m_pending_ops	= node;

	/*
	 * if we got here we must be good
	 */ 
	result =  SW_OKAY;
	
exit:

	if (result != SW_OKAY)
	{
		if (node != NULL)
		{
			sw_free(node);
		}
	}

	sw_salt_unlock(self->m_salt);
	
	return result;
}


sw_result
sw_mdns_stub_stop_browse_domains(
							sw_mdns_stub				self,
							sw_discovery_browse_id	id)
{
	static sw_string	op			=	"stop_browse_domains";
	static sw_ulong	op_len	=	20;
	sw_corby_buffer	buffer;
	sw_result			result;

	sw_salt_lock(self->m_salt);

	/*
	 * try and bind to service
	 */
	SW_TRY_GOTO(sw_mdns_stub_bind(self));

	SW_TRY_GOTO(sw_corby_object_start_request(self->m_service, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) id));
	SW_TRY_GOTO(sw_corby_object_send(self->m_service, buffer, NULL, NULL, NULL));

	result = SW_OKAY;

exit:

	sw_salt_unlock(self->m_salt);
	
	return result;
}


sw_result
sw_mdns_stub_browse_services(
								sw_mdns_stub						self,
								sw_const_string					type,
								sw_const_string					domain,
								sw_discovery_browse_handler	handler,
								sw_discovery_browse_reply		reply,
								sw_opaque							extra,
								sw_discovery_browse_id		*	id)
{
	static sw_string				op 		=	"browse_services";
	static sw_ulong				op_len	=	16;
	sw_corby_buffer 				buffer;
	sw_result						result;
	sw_mdns_stub_pending_op	node		=	NULL;

	sw_salt_lock(self->m_salt);

	/*
	 * try and bind to service
	 */
	SW_TRY_GOTO(sw_mdns_stub_bind(self));

	/*
	 * try to allocate space
	 */
	SW_TRY_GOTO(sw_malloc(node, sw_mdns_stub_pending_op, sizeof(struct _sw_mdns_stub_pending_op)));

	/*
	 * now do garden variety initialization
	 */
	node->m_handler			=	handler;
	node->m_browse_reply		=	reply;
	node->m_extra				=	extra;
	node->m_id					=	sw_mdns_stub_next_id();
	*id							=	(sw_opaque) node->m_id;

	SW_TRY_GOTO(sw_corby_object_start_request(self->m_service, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, type));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, domain));
	SW_TRY_GOTO(sw_corby_buffer_put_object(buffer, self->m_self));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) node->m_id));
	SW_TRY_GOTO(sw_corby_object_send(self->m_service, buffer, NULL, NULL, NULL));
	
	/*
	 * stash the node
	 */
	node->m_txen			= self->m_pending_ops;
	self->m_pending_ops	= node;

	/*
	 * if we got here we must be good
	 */ 
	result =  SW_OKAY;
	
exit:

	if (result != SW_OKAY)
	{
		if (node != NULL)
		{
			sw_free(node);
		}
	}

	sw_salt_unlock(self->m_salt);
	
	return result;
}


sw_result
sw_mdns_stub_stop_browse_services(
							sw_mdns_stub				self,
							sw_discovery_browse_id	id)
{
	static sw_string	op			=	"stop_browse_services";
	static sw_ulong	op_len	=	21;
	sw_corby_buffer	buffer;
	sw_result			result;

	sw_salt_lock(self->m_salt);
	
	/*
	 * try and bind to service
	 */
	SW_TRY_GOTO(sw_mdns_stub_bind(self));

	SW_TRY_GOTO(sw_corby_object_start_request(self->m_service, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) id));
	SW_TRY_GOTO(sw_corby_object_send(self->m_service, buffer, NULL, NULL, NULL));

	result = SW_OKAY;

exit:

	sw_salt_unlock(self->m_salt);
	
	return result;
}


sw_result
sw_mdns_stub_resolve(
							sw_mdns_stub						self,
							sw_const_string					name,
							sw_const_string					reg_type,
							sw_const_string					domain,
							sw_discovery_resolve_handler	handler,
							sw_discovery_resolve_reply	reply,
							sw_opaque							extra,
							sw_discovery_resolve_id	*	id)
{
	static sw_string				op 		=	"resolve";
	static sw_ulong				op_len	=	8;
	sw_corby_buffer 				buffer;
	sw_result						result;
	sw_mdns_stub_pending_op	node		=	NULL;

	sw_salt_lock(self->m_salt);
	
	/*
	 * try and bind to service
	 */
	SW_TRY_GOTO(sw_mdns_stub_bind(self));

	/*
	 * try to allocate space
	 */
	SW_TRY_GOTO(sw_malloc(node, sw_mdns_stub_pending_op, sizeof(struct _sw_mdns_stub_pending_op)));

	/*
	 * now do garden variety initialization
	 */
	node->m_handler			=	handler;
	node->m_resolve_reply	=	reply;
	node->m_extra				=	extra;
	node->m_id					=	sw_mdns_stub_next_id();
	*id							=	(sw_opaque) node->m_id;

	SW_TRY_GOTO(sw_corby_object_start_request(self->m_service, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, name));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, reg_type));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, domain));
	SW_TRY_GOTO(sw_corby_buffer_put_object(buffer, self->m_self));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, node->m_id));
	SW_TRY_GOTO(sw_corby_object_send(self->m_service, buffer, NULL, NULL, NULL));

	/*
	 * stash the node
	 */
	node->m_txen			= self->m_pending_ops;
	self->m_pending_ops	= node;
	
	/*
	 * if we got here we must be good
	 */ 
	result = SW_OKAY;

exit:

	if (result != SW_OKAY)
	{
		if (node != NULL)
		{
			sw_free(node);
		}
	}

	sw_salt_unlock(self->m_salt);
	
	return result;
}


sw_result
sw_mdns_stub_stop_resolve(
						sw_mdns_stub				self,
						sw_discovery_resolve_id	id)
{
	static sw_string	op			=	"stop_resolve";
	static sw_ulong	op_len	=	13;
	sw_corby_buffer	buffer;
	sw_result			result;

	sw_salt_lock(self->m_salt);
	
	/*
	 * try and bind to service
	 */
	SW_TRY_GOTO(sw_mdns_stub_bind(self));

	SW_TRY_GOTO(sw_corby_object_start_request(self->m_service, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) id));
	SW_TRY_GOTO(sw_corby_object_send(self->m_service, buffer, NULL, NULL, NULL));

	result = SW_OKAY;

exit:

	sw_salt_unlock(self->m_salt);
	
	return result;
}


int
sw_mdns_stub_socket(
						sw_mdns_stub	self)
{
	sw_corby_channel	channel;
	sw_socket			socket;

	if (sw_corby_object_channel(self->m_service, &channel) != SW_OKAY)
	{
		return -1;
	}

	if (channel == NULL)
	{
		return -1;
	}

	socket = sw_corby_channel_socket(channel);

	if (socket == NULL)
	{
		return -1;
	}

	return sw_socket_fd(socket);
}


sw_result
sw_mdns_stub_read_socket(
						sw_mdns_stub	self)
{
	sw_corby_channel	channel;

	if (sw_corby_object_channel(self->m_service, &channel) != SW_OKAY)
	{
		return SW_E_FAIL;
	}

	return sw_corby_orb_read_channel(self->m_orb, channel);
}


static sw_result
sw_mdns_stub_bind(
						sw_mdns_stub	self)
{
	sw_assert(self);
	sw_assert(self->m_orb);
	sw_assert(self->m_service);

	if (self->m_bound == SW_FALSE)
	{
		/*
	 	 * set us up for bi-directionality
		 */
		SW_TRY(sw_corby_orb_register_bidirectional_object(self->m_orb, self->m_service));
		self->m_bound = SW_TRUE;
	}

	return SW_OKAY;
}


static sw_result
sw_mdns_stub_dispatcher(
								sw_corby_servant		servant,
								sw_salt					salt,
								sw_corby_orb			orb,
								sw_corby_channel		channel,
								sw_corby_message		message,
								sw_corby_buffer		buffer,
								sw_const_string		op,
								sw_ulong					op_len,
								sw_ulong					request_id,
								sw_octet					endian)
{
	sw_mdns_stub self = (sw_mdns_stub) servant;

	if (sw_strcmp("publish_domain_reply", op) == 0)
	{
		sw_octet						status;
		sw_mdns_stub_pending_op	node;
		sw_ulong						id;
		
		SW_TRY(sw_corby_buffer_get_ulong(buffer, &id, endian));
		SW_TRY(sw_corby_buffer_get_octet(buffer, &status));

		node = sw_mdns_stub_lookup(self, id);

		if (node == NULL)
		{
			sw_debug("unable to lookup pending op: %d\n", id);
			return SW_E_FAIL;
		}
		
		(node->m_publish_domain_reply)(node->m_handler, self->m_discovery, (sw_discovery_publish_status) status, (sw_opaque) id, node->m_extra);
	}
	else if (sw_strcmp("publish_host_reply", op) == 0)
	{
		sw_octet							status;
		sw_mdns_stub_pending_op		node;
		sw_ulong							id;
		
		SW_TRY(sw_corby_buffer_get_ulong(buffer, &id, endian));
		SW_TRY(sw_corby_buffer_get_octet(buffer, &status));

		node = sw_mdns_stub_lookup(self, id);

		if (node == NULL)
		{
			sw_debug("unable to lookup pending op: %d\n", id);
			return SW_E_FAIL;
		}
		
		(node->m_publish_host_reply)(node->m_handler, self->m_discovery, (sw_discovery_publish_status) status, (sw_opaque) id, node->m_extra);
	}
	else if (sw_strcmp("publish_reply", op) == 0)
	{
		sw_octet							status;
		sw_mdns_stub_pending_op		node;
		sw_ulong							id;
		
		SW_TRY(sw_corby_buffer_get_ulong(buffer, &id, endian));
		SW_TRY(sw_corby_buffer_get_octet(buffer, &status));

		node = sw_mdns_stub_lookup(self, id);

		if (node == NULL)
		{
			sw_debug("unable to lookup pending op: %d\n", id);
			return SW_E_FAIL;
		}
		
		(node->m_publish_reply)(node->m_handler, self->m_discovery, (sw_discovery_publish_status) status, (sw_opaque) id, node->m_extra);
	}
	else if (sw_strcmp("browse_reply", op) == 0)
	{
		sw_octet							status;
		sw_mdns_stub_pending_op		node;
		sw_string						name;
		sw_string						type;
		sw_string						domain;
		sw_ulong							dummy;
		sw_ulong							id;
		
		SW_TRY(sw_corby_buffer_get_ulong(buffer, &id, endian));
		SW_TRY(sw_corby_buffer_get_octet(buffer, &status));
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &name, &dummy, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &type, &dummy, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &domain, &dummy, endian));

		node = sw_mdns_stub_lookup(self, id);
		
		if (node == NULL)
		{
			sw_debug("unable to lookup pending op: %d\n", id);
			return SW_E_FAIL;
		}
		
		(node->m_browse_reply)(node->m_handler, self->m_discovery, (sw_opaque) id, (sw_discovery_browse_status) status, name, type, domain, node->m_extra);
	}
	else if (sw_strcmp("resolve_reply", op) == 0)
	{
		sw_string						name;
		sw_string						type;
		sw_string						domain;
		sw_ipv4_address				address;
		sw_port							port;
		sw_string						text_record;
		sw_octets						raw_text_record;
		sw_ulong							raw_text_record_len;
		sw_ulong							dummy;
		sw_mdns_stub_pending_op		node;
		sw_ulong							id;
		
		SW_TRY(sw_corby_buffer_get_ulong(buffer, &id, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &name, &dummy, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &type, &dummy, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &domain, &dummy, endian));
		SW_TRY(sw_corby_buffer_get_ulong(buffer, &dummy, endian));
		SW_TRY(sw_ipv4_address_init_from_saddr(&address, dummy));
		SW_TRY(sw_corby_buffer_get_ushort(buffer, &port, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &text_record, &dummy, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_seq(buffer, &raw_text_record, &raw_text_record_len, endian));

		node = sw_mdns_stub_lookup(self, id);
		
		if (node == NULL)
		{
			sw_debug("unable to lookup pending op: %d\n", id);
			return SW_E_FAIL;
		}
		
		(node->m_resolve_reply)(node->m_handler, self->m_discovery, (sw_opaque) id, name, type, domain, address, port, text_record, raw_text_record, raw_text_record_len, node->m_extra);
	}
	else if (sw_strcmp("check_version_reply", op) == 0)
	{
		sw_octet major;
		sw_octet minor;
		sw_octet micro;

		SW_TRY(sw_corby_buffer_get_ulong(buffer, &self->m_check_version_result, endian));
		SW_TRY(sw_corby_buffer_get_octet(buffer, &major));
		SW_TRY(sw_corby_buffer_get_octet(buffer, &minor));
		SW_TRY(sw_corby_buffer_get_octet(buffer, &micro));

		if (self->m_check_version_result != SW_OKAY)
		{
			sw_log(SW_LOG_WARNING, "sw_mdns_stub_dispatcher", "mdns version mismatch. server version is %d.%d.%d\n", (int) major, (int) minor, (int) micro);
		}
	}
	else
	{
		return SW_E_CORBY_BAD_OPERATION;
	}
	
	return SW_OKAY;
}


static sw_mdns_stub_pending_op
sw_mdns_stub_lookup(
							sw_mdns_stub	self,
							sw_ulong			id)
{
	sw_mdns_stub_pending_op	node;
	
	for (node = self->m_pending_ops; node; node = node->m_txen)
	{
		if (node->m_id == id)
		{
			return node;
		}
	}
	
	return NULL;
}


sw_result
sw_mdns_servant_init(
				sw_mdns_servant	self)
{
	sw_char		scratch[256];
	DNSStatus	status;
	
	/*
	 * only want to bind to loopback so only clients on our own machine can connect.  Use port 5353
	 * just 'cause that's the one mDNS uses.  It could be anything, just needs to be well known.
	 */
	static sw_corby_orb_config	config[] =      
   {
      { "iiop", SW_TAG_INTERNET_IOP,  "127.0.0.1", 5353 },
      { NULL }
   };

	self->m_clients	=	NULL;

	/*
	 * set the port
	 */
	config[0].m_port = self->m_port;

	/*
	 * initialize salt
	 */
	SW_TRY(sw_salt_init(&self->m_salt, 0, NULL));

	/*
	 * now initialize the orb
	 */
	SW_TRY(sw_corby_orb_init(&self->m_orb, self->m_salt, config, self, sw_mdns_servant_connection_notifier, NULL));

	/*
	 * get my port number
	 */
	if (sw_corby_orb_protocol_to_address(self->m_orb, "iiop", scratch, &self->m_port) != SW_OKAY)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_init", "sw_corby_orb_protocol failed\n");
		return SW_E_FAIL;
	}

	/*
	 * register ourselves with the orb
	 */
	SW_TRY(sw_corby_orb_register_servant(self->m_orb, self, (sw_corby_servant_cb) sw_mdns_servant_dispatcher, "dns-sd", NULL, NULL));

	if ((status = DNSServicesInitialize(self->m_salt, kDNSFlagAdvertise, CACHE_SIZE, self->m_filters, self->m_num_filters)) != kDNSNoErr)
	{
		sw_log(SW_LOG_FATAL_ERROR, "sw_mdns_servant_init", "DNSServicesInitialize() failed\n");
		return SW_E_INIT;
	}

	return SW_OKAY;
}


sw_result
sw_mdns_servant_fina(
						sw_mdns_servant	self)
{
	sw_corby_orb_fina(self->m_orb);
	sw_salt_fina(self->m_salt);

	return SW_OKAY;
}


sw_result
sw_mdns_servant_load_file(
						sw_mdns_servant	self,
						sw_const_string	file)
{
	sw_char		line[1024];
	sw_ulong		len;
	FILE		*	f = fopen(file, "r");

	if (f == NULL)
	{
		return SW_E_FAIL;
	}

	while (sw_mdns_servant_getline(self, f, line, &len) == SW_OKAY)
	{
		if (line[0] != '#')
		{
			sw_char 			name[256];
			sw_char 			type[256];
			sw_char 			domain[256];
			sw_ushort		port;
			sw_text_record	text_record;
	
			if (sw_mdns_servant_parse_conf_line(self, line, len, name, type, domain, &port, &text_record) == SW_OKAY)
			{
				DNSStatus				status;
				DNSRegistrationRef	ref;

				status = DNSRegistrationCreate(
									kDNSRegistrationFlagPreFormattedTextRecord|kDNSRegistrationFlagAutoRenameOnConflict,
									name,
									type,
									domain,
									port,
									sw_text_record_bytes(text_record),
									sw_text_record_len(text_record),
									NULL,
									NULL,
									(DNSRegistrationCallBack) sw_mdns_servant_publish_callback2,
									NULL,
									&ref);

				sw_text_record_fina(text_record);
			}
		}
	}

	fclose(f);

	return SW_OKAY;
}


sw_result
sw_mdns_servant_check_version(
						sw_mdns_servant			self,
						const sw_corby_object	client,
						sw_corby_channel			channel,
						sw_octet						major,
						sw_octet						minor,
						sw_octet						micro)
{
	static sw_string					op 		=	"check_version_reply";
	static sw_ulong					op_len	=	20;
	sw_corby_buffer 					buffer;
	sw_mdns_servant_client_node	node;

	SW_TRY(sw_malloc(node, sw_mdns_servant_client_node, sizeof(struct _sw_mdns_servant_client_node)));
	sw_memset(node, 0, sizeof(struct _sw_mdns_servant_client_node));
	sw_mdns_servant_add_node(self, node);

	SW_TRY(sw_corby_object_start_request(client, op, op_len, SW_FALSE, &buffer));

	/*
	 * now check
	 */
	if ((major == SW_MDNS_MAJOR) && (minor == SW_MDNS_MINOR))
	{
		SW_TRY(sw_corby_buffer_put_ulong(buffer, SW_OKAY));
	}
	else
	{
		SW_TRY(sw_corby_buffer_put_ulong(buffer, SW_E_FAIL));
	}

	SW_TRY(sw_corby_buffer_put_octet(buffer, SW_MDNS_MAJOR));
	SW_TRY(sw_corby_buffer_put_octet(buffer, SW_MDNS_MINOR));
	SW_TRY(sw_corby_buffer_put_octet(buffer, SW_MDNS_MICRO));

	return sw_corby_object_send(client, buffer, NULL, NULL, NULL);
}


sw_result
sw_mdns_servant_publish_domain(
						sw_mdns_servant						self,
						sw_const_string						name,
						const sw_corby_object				client,
						sw_corby_channel						channel,
						sw_discovery_publish_domain_id	id)
{
	DNSStatus						status;
	sw_mdns_servant_client_node	node;

	SW_TRY(sw_malloc(node, sw_mdns_servant_client_node, sizeof(struct _sw_mdns_servant_client_node)));
	sw_memset(node, 0, sizeof(struct _sw_mdns_servant_client_node));
	sw_mdns_servant_add_node(self, node);

	/*
	 * now do garden variety initialization
	 */
	node->m_handler					=	self;
	node->m_self						=	self;
	node->m_publish_domain_reply	=	sw_mdns_servant_publish_domain_reply;
	node->m_extra						=	client;
	node->m_channel					=	channel;
	node->m_cleanup_internal		=	sw_mdns_servant_stop_publish_domain_cleanup;
	node->m_cleanup					=	sw_mdns_servant_cleanup;
	node->m_id							=	(sw_ulong) id;
	
	status = DNSDomainRegistrationCreate(
						kDNSDomainRegistrationFlagNone,
						name,
						kDNSDomainRegistrationTypeRegistration,
						&node->m_domain_registration_ref);

	if (status != kDNSNoErr)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_publish", "DNSRegistrationCreate() failed: %d\n", status);
		goto exit;
	}

exit:

	return sw_mdns_servant_convert_status_to_result(status);
}


sw_result
sw_mdns_servant_stop_publish_domain(
						sw_mdns_servant						self,
						sw_corby_channel						channel,
						sw_discovery_publish_domain_id	id)
{
	sw_mdns_servant_client_node node;

	sw_assert(self);
	
	sw_log(SW_LOG_INFO, "sw_mdns_servant_stop_publish_domain_internal", "looking up client %d\n", id);

	node = sw_mdns_servant_lookup_node(self, channel, (sw_ulong) id);

	if (node == NULL)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_stop_publish_internal", "couldn't find client %d\n", id);
		return SW_E_FAIL;
	}

	return sw_mdns_servant_stop_publish_domain_cleanup(self, node);
}


static sw_result
sw_mdns_servant_stop_publish_domain_cleanup(
								sw_mdns_servant		self,
								sw_mdns_servant_client_node	node)
{
	/*
	 * we cleanup the client node in the callback function
	 */
	DNSDomainRegistrationRelease(node->m_domain_registration_ref, 0);

	if (node->m_cleanup != NULL)
	{
		(*node->m_cleanup)(self, node->m_extra);
	}

	sw_mdns_servant_free_node(self, node);

	return SW_OKAY;
}


sw_result
sw_mdns_servant_publish_host(
						sw_mdns_servant						self,
						sw_const_string						name,
						sw_const_string						domain,
						sw_ipv4_address						address,
						const sw_corby_object				client,
						sw_corby_channel						channel,
						sw_discovery_publish_domain_id	id)
{
	DNSStatus						status;
	DNSNetworkAddress				addr;
	sw_mdns_servant_client_node	node;

	SW_TRY(sw_malloc(node, sw_mdns_servant_client_node, sizeof(struct _sw_mdns_servant_client_node)));
	sw_memset(node, 0, sizeof(struct _sw_mdns_servant_client_node));
	sw_mdns_servant_add_node(self, node);

	/*
	 * now do garden variety initialization
	 */
	node->m_handler				=	self;
	node->m_self					=	self;
	node->m_publish_host_reply	=	sw_mdns_servant_publish_host_reply;
	node->m_extra					=	client;
	node->m_channel				=	channel;
	node->m_cleanup_internal	=	sw_mdns_servant_stop_publish_host_cleanup;
	node->m_cleanup				=	sw_mdns_servant_cleanup;
	node->m_id						=	(sw_ulong) id;
	
	addr.addressType				=	kDNSNetworkAddressTypeIPv4;
	addr.u.ipv4.addr.v32			=	sw_ipv4_address_saddr(address);

	status = DNSHostRegistrationCreate(
						kDNSHostRegistrationFlagNone,
						name,
						NULL,
						&addr,
						NULL,
						sw_mdns_servant_publish_host_callback,
						node,
						&node->m_host_registration_ref);
						
	if (status != kDNSNoErr)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_publish_host", "DNSHostRegistrationCreate() failed: %d\n", status);
		goto exit;
	}

exit:

	return sw_mdns_servant_convert_status_to_result(status);
}


sw_result
sw_mdns_servant_stop_publish_host(
						sw_mdns_servant					self,
						sw_corby_channel					channel,
						sw_discovery_publish_host_id	id)
{
	sw_mdns_servant_client_node node;

	sw_assert(self);
	
	sw_log(SW_LOG_INFO, "sw_mdns_servant_stop_publish_host", "looking up client %d\n", id);

	node = sw_mdns_servant_lookup_node(self, channel, (sw_ulong) id);

	if (node == NULL)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_stop_publish_host", "couldn't find client %d\n", id);
		return SW_E_FAIL;
	}

	return sw_mdns_servant_stop_publish_host_cleanup(self, node);
}


static sw_result
sw_mdns_servant_stop_publish_host_cleanup(
								sw_mdns_servant		self,
								sw_mdns_servant_client_node	node)
{
	sw_log(SW_LOG_INFO, "sw_mdns_servant_stop_publish_host_cleanup()", "cleaning up %d\n", node->m_id);

	/*
	 * we cleanup the client node in the callback function
	 */
	DNSHostRegistrationRelease(node->m_host_registration_ref, 0);

	if (node->m_cleanup != NULL)
	{
		(*node->m_cleanup)(self, node->m_extra);
	}

	sw_mdns_servant_free_node(self, node);

	return SW_OKAY;
}


sw_result
sw_mdns_servant_publish(
						sw_mdns_servant				self,
						sw_const_string				name,
						sw_const_string				type,
						sw_const_string				domain,
						sw_const_string				host,
						sw_port							port,
						sw_octets						text_record,
						sw_ulong							text_record_len,
						const sw_corby_object		client,
						sw_corby_channel				channel,
						sw_discovery_publish_id	id)
{
	DNSStatus						status;
	sw_mdns_servant_client_node	node;

	SW_TRY(sw_malloc(node, sw_mdns_servant_client_node, sizeof(struct _sw_mdns_servant_client_node)));
	sw_memset(node, 0, sizeof(struct _sw_mdns_servant_client_node));
	sw_mdns_servant_add_node(self, node);

	/*
	 * now do garden variety initialization
	 */
	node->m_handler				=	self;
	node->m_self					=	self;
	node->m_publish_reply		=	sw_mdns_servant_publish_reply;
	node->m_extra					=	client;
	node->m_channel				=	channel;
	node->m_cleanup_internal	=	sw_mdns_servant_stop_publish_cleanup;
	node->m_cleanup				=	sw_mdns_servant_cleanup;
	node->m_id						=	(sw_ulong) id;
	
	status = DNSRegistrationCreate(
						kDNSRegistrationFlagPreFormattedTextRecord|kDNSRegistrationFlagAutoRenameOnConflict,
						name,
						type,
						domain,
						port,
						text_record,
						text_record_len,
						host,
						NULL,
						(DNSRegistrationCallBack) sw_mdns_servant_publish_callback,
						node,
						&node->m_registration_ref);

	if (status != kDNSNoErr)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_publish", "DNSRegistrationCreate() failed: %d\n", status);
		goto exit;
	}

exit:

	return sw_mdns_servant_convert_status_to_result(status);
}


sw_result
sw_mdns_servant_publish_update(
						sw_mdns_servant				self,
						sw_corby_channel				channel,
						sw_discovery_publish_id	id,
						sw_octets						text_record,
						sw_ulong							text_record_len)
{
	sw_mdns_servant_client_node node;

	sw_assert(self);
	
	sw_log(SW_LOG_INFO, "sw_mdns_servant_publish_update", "looking up client %d\n", id);

	node = sw_mdns_servant_lookup_node(self, channel, (sw_ulong) id);

	if (node == NULL)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_publish_update", "couldn't find client %d\n", id);
		return SW_E_FAIL;
	}

	DNSRegistrationUpdate(node->m_registration_ref,	0, NULL, text_record, text_record_len, 120);

	return SW_OKAY;
}


sw_result
sw_mdns_servant_stop_publish(
						sw_mdns_servant				self,
						sw_corby_channel				channel,
						sw_discovery_publish_id	id)
{
	sw_mdns_servant_client_node node;

	sw_assert(self);
	
	sw_log(SW_LOG_INFO, "sw_mdns_servant_stop_publish", "looking up client %d\n", id);

	node = sw_mdns_servant_lookup_node(self, channel, (sw_ulong) id);

	if (node == NULL)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_stop_publish", "couldn't find client %d\n", id);
		return SW_E_FAIL;
	}

	return sw_mdns_servant_stop_publish_cleanup(self, node);
}


static sw_result
sw_mdns_servant_stop_publish_cleanup(
								sw_mdns_servant		self,
								sw_mdns_servant_client_node	node)
{
	/*
	 * we cleanup the client node in the callback function
	 */
	DNSRegistrationRelease(node->m_registration_ref, 0);

	if (node->m_cleanup != NULL)
	{
		(*node->m_cleanup)(self, node->m_extra);
	}

	sw_mdns_servant_free_node(self, node);

	return SW_OKAY;
}


sw_result
sw_mdns_servant_browse_domains(
						sw_mdns_servant			self,
						const sw_corby_object	client,
						sw_corby_channel			channel,
						sw_discovery_browse_id	id)
{
	DNSStatus						status;
	sw_result						result;
	sw_mdns_servant_client_node	node;
	
	/*
	 * try to allocate space
	 */
	SW_TRY(sw_malloc(node, sw_mdns_servant_client_node, sizeof(struct _sw_mdns_servant_client_node)));
	sw_memset(node, 0, sizeof(struct _sw_mdns_servant_client_node));
	sw_mdns_servant_add_node(self, node);

	/*
	 * now do garden variety initialization
	 */
	node->m_handler				= 	self;
	node->m_self					=	self;
	node->m_browse_reply			= 	sw_mdns_servant_browse_reply;
	node->m_extra					=	client;
	node->m_channel				=	channel;
	node->m_cleanup_internal	=	sw_mdns_servant_stop_browse_domains_cleanup;
	node->m_cleanup				=	sw_mdns_servant_cleanup;
	node->m_id						=	(sw_ulong) id;

	status = DNSBrowserCreate(0, (DNSBrowserCallBack) sw_mdns_servant_browse_callback, node, &node->m_browser_ref);

	if (status != kDNSNoErr)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_browse_domains", "DNSBrowserCreate() failed: %d\n", status);
		goto exit;
	}
	
	status = DNSBrowserStartDomainSearch(node->m_browser_ref, 0);
	
	if (status != kDNSNoErr)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_browse_domains", "DNSBrowserStartDomainSearch() failed: %d\n", status);
		goto exit;
	}
	
exit:

	return sw_mdns_servant_convert_status_to_result(status);
	
	return result;
}


sw_result
sw_mdns_servant_stop_browse_domains(
						sw_mdns_servant			self,
						sw_corby_channel			channel,
						sw_discovery_browse_id	id)
{
	sw_mdns_servant_client_node	node;

	sw_assert(self);
	
	sw_log(SW_LOG_INFO, "sw_mdns_servant_stop_browse_domains", "looking up client %d\n", id);

	node = sw_mdns_servant_lookup_node(self, channel, (sw_ulong) id);

	if (node == NULL)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_stop_browse_domains", "couldn't find client %d\n", id);
		return SW_E_FAIL;
	}

	return sw_mdns_servant_stop_browse_domains_cleanup(self, node);
}


static sw_result
sw_mdns_servant_stop_browse_domains_cleanup(
									sw_mdns_servant		self,
									sw_mdns_servant_client_node	node)
{
	DNSBrowserRelease(node->m_browser_ref, 0);

	if (node->m_cleanup != NULL)
	{
		(*node->m_cleanup)(self, node->m_extra);
	}

	sw_mdns_servant_free_node(self, node);

	return SW_OKAY;
}


sw_result
sw_mdns_servant_browse_services(
						sw_mdns_servant	self,
						sw_const_string			type,
						sw_const_string			domain,
						const sw_corby_object	client,
						sw_corby_channel			channel,
						sw_discovery_browse_id	id)
{
	DNSStatus						status;
	sw_mdns_servant_client_node	node;
	
	/*
	 * try to allocate space
	 */
	SW_TRY(sw_malloc(node, sw_mdns_servant_client_node, sizeof(struct _sw_mdns_servant_client_node)));
	sw_memset(node, 0, sizeof(struct _sw_mdns_servant_client_node));
	sw_mdns_servant_add_node(self, node);

	/*
	 * now do garden variety initialization
	 */
	node->m_handler				= 	self;
	node->m_self					=	self;
	node->m_browse_reply			= 	sw_mdns_servant_browse_reply;
	node->m_extra					=	client;
	node->m_channel				=	channel;
	node->m_cleanup_internal	=	sw_mdns_servant_stop_browse_services_cleanup;
	node->m_cleanup				=	sw_mdns_servant_cleanup;
	node->m_id						=	(sw_ulong) id;

	status = DNSBrowserCreate(0, (DNSBrowserCallBack) sw_mdns_servant_browse_callback, node, &node->m_browser_ref);

	if (status != kDNSNoErr)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_browse_services", "DNSBrowserCreate() failed: %d\n", status);
		goto exit;
	}
	
	status = DNSBrowserStartServiceSearch(node->m_browser_ref, 0, type, domain);
	
	if (status != kDNSNoErr)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_browse_services", "DNSBrowserStartServiceSearch() failed: %d\n", status);
		goto exit;
	}
	
exit:

	return sw_mdns_servant_convert_status_to_result(status);
}


sw_result
sw_mdns_servant_stop_browse_services(
						sw_mdns_servant	self,
						sw_corby_channel			channel,
						sw_discovery_browse_id	id)
{
	sw_mdns_servant_client_node node;

	sw_assert(self);
	
	sw_log(SW_LOG_INFO, "sw_mdns_servant_stop_browse_services", "looking up client %d\n", id);

	node = sw_mdns_servant_lookup_node(self, channel, (sw_ulong) id);

	if (node == NULL)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_stop_browse_services", "couldn't find client %d\n", id);
		return SW_E_FAIL;
	}

	return sw_mdns_servant_stop_browse_services_cleanup(self, node);
}


static sw_result
sw_mdns_servant_stop_browse_services_cleanup(
							sw_mdns_servant		self,
							sw_mdns_servant_client_node	node)
{
	DNSBrowserRelease(node->m_browser_ref, 0);

	if (node->m_cleanup)
	{
		(*node->m_cleanup)(self, node->m_extra);
	}

	sw_mdns_servant_free_node(self, node);

	return SW_OKAY;
}


sw_result
sw_mdns_servant_resolve(
						sw_mdns_servant		self,
						sw_const_string				name,
						sw_const_string				type,
						sw_const_string				domain,
						const sw_corby_object		client,
						sw_corby_channel				channel,
						sw_discovery_resolve_id	id)
{
	DNSStatus						status;
	sw_mdns_servant_client_node	node;
	
	/*
	 * try to allocate space
	 */
	SW_TRY(sw_malloc(node, sw_mdns_servant_client_node, sizeof(struct _sw_mdns_servant_client_node)));
	sw_memset(node, 0, sizeof(struct _sw_mdns_servant_client_node));
	sw_mdns_servant_add_node(self, node);

	/*
	 * now do garden variety initialization
	 */
	node->m_handler				= 	self;
	node->m_self					=	self;
	node->m_resolve_reply		= 	sw_mdns_servant_resolve_reply;
	node->m_extra					=	client;
	node->m_channel				=	channel;
	node->m_cleanup_internal	=	sw_mdns_servant_stop_resolve_cleanup;
	node->m_cleanup				=	sw_mdns_servant_cleanup;
	node->m_id						=	(sw_ulong) id;
	
	status = DNSResolverCreate(0, name, type, domain, (DNSResolverCallBack) sw_mdns_servant_resolve_callback, node, NULL, &node->m_resolver_ref);

	if (status != kDNSNoErr)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_resolve", "DNSResolverCreate() failed: %d\n", status);
		goto exit;
	}
	
exit:

	return sw_mdns_servant_convert_status_to_result(status);
}


sw_result
sw_mdns_servant_stop_resolve(
						sw_mdns_servant		self,
						sw_corby_channel				channel,
						sw_discovery_resolve_id	id)
{
	sw_mdns_servant_client_node node;

	sw_log(SW_LOG_INFO, "sw_mdns_servant_stop_resolve", "looking up client %d\n", id);
	
	node = sw_mdns_servant_lookup_node(self, channel, (sw_ulong) id);

	if (node == NULL)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_stop_resolve", "couldn't find client %d\n", id);
		return SW_E_FAIL;
	}

	return sw_mdns_servant_stop_resolve_cleanup(self, node);
}


static sw_result
sw_mdns_servant_stop_resolve_cleanup(
							sw_mdns_servant		self,
							sw_mdns_servant_client_node	node)
{
	DNSResolverRelease(node->m_resolver_ref, 0);

	if (node->m_cleanup)
	{
		(*node->m_cleanup)(self, node->m_extra);
	}

	sw_mdns_servant_free_node(self, node);

	return SW_OKAY;
}


static sw_result
sw_mdns_servant_dispatcher(
            sw_corby_servant	servant,
            sw_salt				salt,
            sw_corby_orb		orb,
            sw_corby_channel  channel,
            sw_corby_message  message,
            sw_corby_buffer   buffer,
            sw_const_string   op,
            sw_ulong          op_len,
            sw_ulong				request_id,
            sw_octet				endian)
{
	sw_mdns_servant self = (sw_mdns_servant) servant;

	if (sw_strcmp("publish_domain", op) == 0)
	{
		sw_ulong							len;
		sw_string						name;
		sw_corby_object				client;
		sw_discovery_publish_id	id;
		
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &name, &len, endian));
		SW_TRY(sw_corby_buffer_get_object(buffer, &client, endian));
		SW_TRY(sw_corby_buffer_get_ulong(buffer, (sw_ulong*) &id, endian));
		SW_TRY(sw_corby_object_set_channel(client, channel));
		
		if (sw_mdns_servant_publish_domain(self, name, client, channel, id) != SW_OKAY)
		{
			return sw_mdns_servant_publish_reply(self, NULL, SW_DISCOVERY_PUBLISH_INVALID, id, client);
		}
	}
	else if (sw_strcmp("stop_publish_domain", op) == 0)
	{
		sw_discovery_publish_domain_id id;
		
		SW_TRY(sw_corby_buffer_get_ulong(buffer, (sw_ulong*) &id, endian));
		SW_TRY(sw_mdns_servant_stop_publish_domain(self, channel, id));
	}
	else if (sw_strcmp("publish_host", op) == 0)
	{
		sw_ulong							len;
		sw_string						name;
		sw_string						domain;
		sw_ulong							saddr;
		sw_ipv4_address				address;
		sw_corby_object				client;
		sw_discovery_publish_id	id;
		
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &name, &len, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &domain, &len, endian));
		SW_TRY(sw_corby_buffer_get_ulong(buffer, &saddr, endian));
		SW_TRY(sw_ipv4_address_init_from_saddr(&address, saddr));
		SW_TRY(sw_corby_buffer_get_object(buffer, &client, endian));
		SW_TRY(sw_corby_buffer_get_ulong(buffer, (sw_ulong*) &id, endian));
		SW_TRY(sw_corby_object_set_channel(client, channel));
		
		if (sw_mdns_servant_publish_host(self, name, domain, address, client, channel, id) != SW_OKAY)
		{
			return sw_mdns_servant_publish_reply(self, NULL, SW_DISCOVERY_PUBLISH_INVALID, id, client);
		}
	}
	else if (sw_strcmp("stop_publish_host", op) == 0)
	{
		sw_discovery_publish_host_id id;
		
		SW_TRY(sw_corby_buffer_get_ulong(buffer, (sw_ulong*) &id, endian));
		SW_TRY(sw_mdns_servant_stop_publish_host(self, channel, id));
	}
	else if (sw_strcmp("publish", op) == 0)
	{
		sw_ulong							len;
		sw_string						name;
		sw_string						type;
		sw_string						domain;
		sw_string						host;
		sw_port							port;
		sw_octets						text_record;
		sw_ulong							text_record_len;
		sw_corby_object				client;
		sw_discovery_publish_id	id;
		
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &name, &len, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &type, &len, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &domain, &len, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &host, &len, endian));
		SW_TRY(sw_corby_buffer_get_ushort(buffer, &port, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_seq(buffer, &text_record, &text_record_len, endian));
		SW_TRY(sw_corby_buffer_get_object(buffer, &client, endian));
		SW_TRY(sw_corby_buffer_get_ulong(buffer, (sw_ulong*) &id, endian));
		SW_TRY(sw_corby_object_set_channel(client, channel));
		
		if (sw_mdns_servant_publish(self, name, type, domain, host, port, text_record, text_record_len, client, channel, id) != SW_OKAY)
		{
			return sw_mdns_servant_publish_reply(self, NULL, SW_DISCOVERY_PUBLISH_INVALID, id, client);
		}
	}
	else if (sw_strcmp("publish_update", op) == 0)
	{
		sw_discovery_publish_id	id;
		sw_octets						text_record;
		sw_ulong							text_record_len;

		SW_TRY(sw_corby_buffer_get_ulong(buffer, (sw_ulong*) &id, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_seq(buffer, &text_record, &text_record_len, endian));
		SW_TRY(sw_mdns_servant_publish_update(self, channel, id, text_record, text_record_len));
	}
	else if (sw_strcmp("stop_publish", op) == 0)
	{
		sw_discovery_publish_id id;
		
		SW_TRY(sw_corby_buffer_get_ulong(buffer, (sw_ulong*) &id, endian));
		SW_TRY(sw_mdns_servant_stop_publish(self, channel, id));
	}
	else if (sw_strcmp("browse_domains", op) == 0)
	{
		sw_corby_object			client;
		sw_discovery_browse_id	id;
		
		SW_TRY(sw_corby_buffer_get_object(buffer, &client, endian));
		SW_TRY(sw_corby_buffer_get_ulong(buffer, (sw_ulong*) &id, endian));
		SW_TRY(sw_corby_object_set_channel(client, channel));
		
		if (sw_mdns_servant_browse_domains(self, client, channel, id) != SW_OKAY)
		{
			return sw_mdns_servant_browse_reply(self, NULL, id, SW_DISCOVERY_BROWSE_INVALID, NULL, NULL, NULL, client);
		}
	}
	else if (sw_strcmp("stop_browse_domains", op) == 0)
	{
		sw_discovery_browse_id id;
		
		SW_TRY(sw_corby_buffer_get_ulong(buffer, (sw_ulong*) &id, endian));
		SW_TRY(sw_mdns_servant_stop_browse_domains(self, channel, id));
	}
	else if (sw_strcmp("browse_services", op) == 0)
	{
		sw_ulong						len;
		sw_string					type;
		sw_string					domain;
		sw_corby_object			client;
		sw_discovery_browse_id	id;
		
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &type, &len, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &domain, &len, endian));
		SW_TRY(sw_corby_buffer_get_object(buffer, &client, endian));
		SW_TRY(sw_corby_buffer_get_ulong(buffer, (sw_ulong*) &id, endian));
		SW_TRY(sw_corby_object_set_channel(client, channel));
		
		if (sw_mdns_servant_browse_services(self, type, domain, client, channel, id) != SW_OKAY)
		{
			return sw_mdns_servant_browse_reply(self, NULL, id, SW_DISCOVERY_BROWSE_INVALID, NULL, NULL, NULL, client);
		}
	}
	else if (sw_strcmp("stop_browse_services", op) == 0)
	{
		sw_discovery_browse_id id;
		
		SW_TRY(sw_corby_buffer_get_ulong(buffer, (sw_ulong*) &id, endian));
		SW_TRY(sw_mdns_servant_stop_browse_services(self, channel, id));
	}
	else if (sw_strcmp("resolve", op) == 0)
	{
		sw_ulong							len;
		sw_string						name;
		sw_string						type;
		sw_string						domain;
		sw_corby_object				client;
		sw_discovery_resolve_id	id;
		
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &name, &len, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &type, &len, endian));
		SW_TRY(sw_corby_buffer_get_zerocopy_cstring(buffer, &domain, &len, endian));
		SW_TRY(sw_corby_buffer_get_object(buffer, &client, endian));
		SW_TRY(sw_corby_buffer_get_ulong(buffer, (sw_ulong*) &id, endian));
		SW_TRY(sw_corby_object_set_channel(client, channel));
		SW_TRY(sw_mdns_servant_resolve(self, name, type, domain, client, channel, id));
	}
	else if (sw_strcmp("stop_resolve", op) == 0)
	{
		sw_discovery_resolve_id id;
		
		SW_TRY(sw_corby_buffer_get_ulong(buffer, (sw_ulong*) &id, endian));
		SW_TRY(sw_mdns_servant_stop_resolve(self, channel, id));
	}
	else if (sw_strcmp("check_version", op) == 0)
	{
		sw_corby_object	client;
		sw_octet 			major;
		sw_octet 			minor;
		sw_octet 			micro;

		SW_TRY(sw_corby_buffer_get_object(buffer, &client, endian));
		SW_TRY(sw_corby_object_set_channel(client, channel));
		SW_TRY(sw_corby_buffer_get_octet(buffer, &major));
		SW_TRY(sw_corby_buffer_get_octet(buffer, &minor));
		SW_TRY(sw_corby_buffer_get_octet(buffer, &micro));

		SW_TRY(sw_mdns_servant_check_version(self, client, channel, major, minor, micro));
	}
	else
	{
		return SW_E_CORBY_BAD_OPERATION;
	}
	
	return SW_OKAY;
}


static sw_result
sw_mdns_servant_connection_notifier(
				sw_corby_orb_observer			handler,
				sw_salt								salt,
				sw_corby_orb						orb,
				struct _sw_corby_channel	*	channel,
				sw_opaque							extra)
{
	sw_mdns_servant servant = (sw_mdns_servant) handler;

	return sw_mdns_servant_cleanup_client(servant, channel);
}


static sw_result
sw_mdns_servant_cleanup(
				sw_mdns_servant			self,
				sw_opaque							extra)
{
	sw_corby_object object = (sw_corby_object) extra;

	sw_corby_object_fina(object);

	return SW_TRUE;
}


static sw_result
sw_mdns_servant_publish_domain_reply(
				sw_discovery_publish_domain_handler	handler,
				sw_discovery									discovery,
				sw_discovery_publish_status				status,
				sw_discovery_publish_domain_id			id,
				sw_opaque										extra)
{
	static sw_string					op			=	"publish_domain_reply";
	static sw_ulong					op_len	=	21;
	sw_mdns_servant			self;
	sw_corby_object					client;
	sw_corby_buffer					buffer;
	sw_result							result;

	sw_debug("sw_mdns_servant_publish_domain_reply\n");

	self		= (sw_mdns_servant) handler;
	client	= (sw_corby_object) extra;
	
	SW_TRY_GOTO(sw_corby_object_start_request(client, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) id));
	SW_TRY_GOTO(sw_corby_buffer_put_octet(buffer, (sw_octet) status));
	SW_TRY_GOTO(sw_corby_object_send(client, buffer, NULL, NULL, NULL));
	
	result = SW_OKAY;
	
exit:

	return result;
};


static sw_result
sw_mdns_servant_publish_host_reply(
				sw_discovery_publish_host_handler	handler,
				sw_discovery								discovery,
				sw_discovery_publish_status			status,
				sw_discovery_publish_host_id			id,
				sw_opaque									extra)
{
	static sw_string					op			=	"publish_host_reply";
	static sw_ulong					op_len	=	19;
	sw_mdns_servant			self;
	sw_corby_object					client;
	sw_corby_buffer					buffer;
	sw_result							result;

	sw_debug("sw_mdns_servant_publish_host_reply\n");

	self		= (sw_mdns_servant) handler;
	client	= (sw_corby_object) extra;
	
	SW_TRY_GOTO(sw_corby_object_start_request(client, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) id));
	SW_TRY_GOTO(sw_corby_buffer_put_octet(buffer, (sw_octet) status));
	SW_TRY_GOTO(sw_corby_object_send(client, buffer, NULL, NULL, NULL));
	
	result = SW_OKAY;
	
exit:

	return result;
}


static sw_result
sw_mdns_servant_publish_reply(
				sw_discovery_publish_handler	handler,
				sw_discovery						discovery,
				sw_discovery_publish_status	status,
				sw_discovery_publish_id		id,
				sw_opaque							extra)
{
	static sw_string					op			=	"publish_reply";
	static sw_ulong					op_len	=	14;
	sw_mdns_servant			self;
	sw_corby_object					client;
	sw_corby_buffer					buffer;
	sw_result							result;

	sw_debug("sw_mdns_servant_publish_reply\n");

	self		= (sw_mdns_servant) handler;
	client	= (sw_corby_object) extra;
	
	SW_TRY_GOTO(sw_corby_object_start_request(client, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) id));
	SW_TRY_GOTO(sw_corby_buffer_put_octet(buffer, (sw_octet) status));
	SW_TRY_GOTO(sw_corby_object_send(client, buffer, NULL, NULL, NULL));
	
	result = SW_OKAY;
	
exit:

	return result;
}


static sw_result
sw_mdns_servant_browse_reply(
				sw_discovery_browse_handler	handler,
				sw_discovery						discovery,
				sw_discovery_browse_id			id,
				sw_discovery_browse_status	status,
				sw_const_string					name,
				sw_const_string					type,
				sw_const_string					domain,
				sw_opaque							extra)
{
	static sw_string				op			=	"browse_reply";
	static sw_ulong				op_len	=	13;
	sw_mdns_servant		self;
	sw_corby_object				client;
	sw_corby_buffer				buffer;
	sw_result						result;

	sw_log(SW_LOG_INFO, "sw_mdns_servant_browse_reply", "%s %s %s\n",
				(name != NULL) ? name : "<null>",
				(type != NULL) ? type : "<null>",
				(domain != NULL) ? domain : "<null>");
	
	self		= (sw_mdns_servant) handler;
	client	= (sw_corby_object) extra;

	SW_TRY_GOTO(sw_corby_object_start_request(client, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) id));
	SW_TRY_GOTO(sw_corby_buffer_put_octet(buffer, (sw_octet) status));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, name));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, type));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, domain));
	SW_TRY_GOTO(sw_corby_object_send(client, buffer, NULL, NULL, NULL));
	
	result = SW_OKAY;

exit:

	return result;
}


static sw_result
sw_mdns_servant_resolve_reply(
				sw_discovery_resolve_handler	handler,
				sw_discovery						discovery,
				sw_discovery_resolve_id		id,
				sw_const_string					name,
				sw_const_string					type,
				sw_const_string					domain,
				sw_ipv4_address					address,
				sw_port								port,
				sw_const_string					text_record_string,
				sw_octets							text_record,
				sw_ulong								text_record_len,
				sw_opaque							extra)
{
	static sw_string				op			=	"resolve_reply";
	static sw_ulong				op_len	=	14;
	sw_mdns_servant		self;
	sw_corby_object				client;
	sw_corby_buffer				buffer;
	sw_result						result;

	sw_log(SW_LOG_INFO, "sw_mdns_servant_resolve_reply", "%s, %s, %s, %d\n",
				(name != NULL) ? name : "<null>",
				(type != NULL) ? type : "<null>",
				(domain != NULL) ? domain : "<null>",
				(int) port);
	
	self		= (sw_mdns_servant) handler;
	client	= (sw_corby_object) extra;

	SW_TRY_GOTO(sw_corby_object_start_request(client, op, op_len, SW_FALSE, &buffer));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, (sw_ulong) id));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, name));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, type));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, domain));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, sw_ipv4_address_saddr(address)));
	SW_TRY_GOTO(sw_corby_buffer_put_ushort(buffer, port));
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(buffer, text_record_string));
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(buffer, text_record_len));
	SW_TRY_GOTO(sw_corby_buffer_put_buf(buffer, text_record, text_record_len));
	SW_TRY_GOTO(sw_corby_object_send(client, buffer, NULL, NULL, NULL));

	result = SW_OKAY;
	
exit:

	if (result != SW_OKAY)
	{
		sw_log(SW_LOG_ERROR, "sw_mdns_servant_resolve_reply", "unable to communicate with client");
	}
	
	return result;
}


static sw_result
sw_mdns_servant_host_registration_reply(
				void *						inContext, 
				DNSHostRegistrationRef 	inRef, 
				DNSStatus 					inStatusCode, 
				void *						inData)
{
}


sw_result
sw_mdns_servant_cleanup_client(
								sw_mdns_servant	self,
								sw_corby_channel			channel)
{
	sw_mdns_servant_client_node node = self->m_clients;

	while (node)
	{
		if (node->m_channel == channel)
		{
			sw_mdns_servant_client_node next = node->m_txen;
			sw_assert(node->m_cleanup_internal);
			(*node->m_cleanup_internal)(self, node);
			node = next;
		}
		else
		{
			node = node->m_txen;
		}
	}

	return SW_OKAY;
}


static void
sw_mdns_servant_publish_host_callback(
			void							*	inContext, 
			DNSHostRegistrationRef		inRef, 
			DNSStatus						inStatusCode, 
			void							*	inData )
{
	sw_mdns_servant_client_node node;
	
	node = (sw_mdns_servant_client_node) inContext;
	
	(node->m_publish_host_reply)(node->m_handler, NULL, (inStatusCode == kDNSNoErr) ? SW_DISCOVERY_PUBLISH_STARTED : SW_DISCOVERY_PUBLISH_INVALID, (sw_opaque) node->m_id, node->m_extra);
}


static sw_result
sw_mdns_servant_publish_callback(
				void								*	inContext,
				DNSRegistrationRef            inRef,
				DNSStatus                  	inStatusCode,
				const DNSRegistrationEvent *  inEvent)
{
	sw_mdns_servant_client_node node;
	
	node = (sw_mdns_servant_client_node) inContext;
	
	if (inStatusCode != kDNSNoErr)
	{
		sw_log(SW_LOG_ERROR, "sw_discovery_publish_callback", "inStatusCode is %d\n", inStatusCode);
		return SW_E_INIT;
	}

	switch (inEvent->type)
	{
		case kDNSRegistrationEventTypeInvalid: 
		{
			(node->m_publish_reply)(node->m_handler, NULL, SW_DISCOVERY_PUBLISH_INVALID, (sw_opaque) node->m_id, node->m_extra);
		}
		break;

		case kDNSRegistrationEventTypeRelease:
		{
			(node->m_publish_reply)(node->m_handler, NULL, SW_DISCOVERY_PUBLISH_STOPPED, (sw_opaque) node->m_id, node->m_extra);
		}
		break;

		case kDNSRegistrationEventTypeRegistered:
		{
			(node->m_publish_reply)(node->m_handler, NULL, SW_DISCOVERY_PUBLISH_STARTED, (sw_opaque) node->m_id, node->m_extra);
		}
		break;

		case kDNSRegistrationEventTypeNameCollision:
		{
			(node->m_publish_reply)(node->m_handler, NULL, SW_DISCOVERY_PUBLISH_NAME_COLLISION, (sw_opaque) node->m_id, node->m_extra);
		}
		break;
	}

	
	return SW_OKAY;
}



static sw_result
sw_mdns_servant_publish_callback2(
				void								*	inContext,
				DNSRegistrationRef            inRef,
				DNSStatus                  	inStatusCode,
				const DNSRegistrationEvent *  inEvent)
{
	if (inStatusCode != kDNSNoErr)
	{
		sw_log(SW_LOG_ERROR, "sw_discovery_publish_callback2", "inStatusCode is %d\n", inStatusCode);
		return SW_E_INIT;
	}

	return SW_OKAY;
}


static sw_result
sw_mdns_servant_browse_callback(
			void							*	inContext,
			DNSBrowserRef					inRef,
			DNSStatus						inStatusCode,
			const DNSBrowserEvent	*	inEvent)
{
	sw_mdns_servant_client_node node;
	
	node = (sw_mdns_servant_client_node) inContext;

	if (inStatusCode != kDNSNoErr)
	{
		sw_log(SW_LOG_ERROR, "sw_discovery_browse_callback", "inStatusCode is %d\n", inStatusCode);
		return SW_E_INIT;
	}
	
	switch (inEvent->type)
	{
		case kDNSBrowserEventTypeInvalid:
		{	
			(node->m_browse_reply)(node->m_handler, NULL, (sw_opaque) node->m_id, SW_DISCOVERY_BROWSE_INVALID, NULL, NULL, NULL, node->m_extra);
		}
		break;
		
		case kDNSBrowserEventTypeRelease:
		{
			(node->m_browse_reply)(node->m_handler, NULL, (sw_opaque) node->m_id, SW_DISCOVERY_BROWSE_RELEASE, NULL, NULL, NULL, node->m_extra);

		}
		break;
		
		case kDNSBrowserEventTypeAddDomain:
		{
			(node->m_browse_reply)(node->m_handler, NULL, (sw_opaque) node->m_id, SW_DISCOVERY_BROWSE_ADD_DOMAIN, NULL, NULL, inEvent->data.addDomain.domain, node->m_extra);
		}
		break;
		
		case kDNSBrowserEventTypeAddDefaultDomain:
		{
			(node->m_browse_reply)(node->m_handler, NULL, (sw_opaque) node->m_id, SW_DISCOVERY_BROWSE_ADD_DEFAULT_DOMAIN, NULL, NULL, inEvent->data.addDomain.domain, node->m_extra);
		}
		break;
		
		case kDNSBrowserEventTypeRemoveDomain:
		{
			(node->m_browse_reply)(node->m_handler, NULL, (sw_opaque) node->m_id, SW_DISCOVERY_BROWSE_REMOVE_DOMAIN, NULL, NULL, inEvent->data.addDomain.domain, node->m_extra);
		}
		break;
		
		case kDNSBrowserEventTypeAddService:
		{
			(node->m_browse_reply)(node->m_handler, NULL, (sw_opaque) node->m_id, SW_DISCOVERY_BROWSE_ADD_SERVICE, inEvent->data.addService.name, inEvent->data.addService.type, inEvent->data.addService.domain, node->m_extra);
		}
		break;
		
		case kDNSBrowserEventTypeRemoveService:
		{
			(node->m_browse_reply)(node->m_handler, NULL, (sw_opaque) node->m_id, SW_DISCOVERY_BROWSE_REMOVE_SERVICE, inEvent->data.removeService.name, inEvent->data.removeService.type, inEvent->data.removeService.domain, node->m_extra);
		}
		break;
		
		case kDNSBrowserEventTypeResolved:
		{
			(node->m_browse_reply)(node->m_handler, NULL, (sw_opaque) node->m_id, SW_DISCOVERY_BROWSE_RESOLVED, inEvent->data.addService.name, inEvent->data.addService.type, inEvent->data.addDomain.domain, node->m_extra);
		}
		break;
	}
	return SW_OKAY;
}


static sw_result
sw_mdns_servant_resolve_callback(
				void								*	inContext,
				DNSResolverRef						inRef,
				DNSStatus                  	inStatusCode,
				const DNSResolverEvent		*	inEvent)
{
	sw_mdns_servant_client_node	node;
	sw_ipv4_address				address;
	char								name_buf[16];
	struct in_addr 				dummy;
	
	node = (sw_mdns_servant_client_node) inContext;

	if (inStatusCode != kDNSNoErr)
	{
		sw_log(SW_LOG_ERROR, "sw_discovery_resolve_callback", "inStatusCode is %d\n", inStatusCode);
		return SW_E_INIT;
	}
	
	/*
	 * ignore events that are not resolve events
	 */
	if (inEvent->type != kDNSResolverEventTypeResolved)
	{
		return SW_OKAY;
	}

	if (inEvent->data.resolved.address.addressType != kDNSNetworkAddressTypeIPv4)
	{
		return SW_OKAY;
	}

	sw_ipv4_address_init_from_saddr(&address, inEvent->data.resolved.address.u.ipv4.addr.v32);

	dummy.s_addr = inEvent->data.resolved.address.u.ipv4.addr.v32;

	sw_log(SW_LOG_INFO, "sw_discovery_resolve_callback", "name %s, address %s %s, port = %d\n", (int) inEvent->data.resolved.name, sw_ipv4_address_name(address, name_buf, 16), inet_ntoa(dummy), ntohs(inEvent->data.resolved.address.u.ipv4.port.v16));

	(node->m_resolve_reply)(node->m_handler, NULL, (sw_opaque) node->m_id, inEvent->data.resolved.name, inEvent->data.resolved.type, inEvent->data.resolved.domain, address, ntohs(inEvent->data.resolved.address.u.ipv4.port.v16), (sw_const_string) inEvent->data.resolved.textRecord, (sw_octets) inEvent->data.resolved.textRecordRaw, inEvent->data.resolved.textRecordRawSize, node->m_extra);

	return SW_OKAY;
}


static sw_result
sw_mdns_servant_add_node(
					sw_mdns_servant					self,
					sw_mdns_servant_client_node	node)
{
	node->m_txen	=	NULL;
	node->m_verp	=	NULL;

	if (self->m_clients)
	{
		self->m_clients->m_verp = node;
	}
	
	node->m_txen		=	self->m_clients;
	self->m_clients	=	node;
	
	return SW_OKAY;
}


static sw_mdns_servant_client_node
sw_mdns_servant_lookup_node(
					sw_mdns_servant	self,
					sw_corby_channel	channel,
					sw_ulong				id)
{
	sw_mdns_servant_client_node node;
	
	for (node = self->m_clients; node; node = node->m_txen)
	{
		if ((node->m_channel == channel) && (node->m_id == id))
		{
			return node;
		}
	}

	return NULL;
}


static sw_result
sw_mdns_servant_free_node(
					sw_mdns_servant					self,
					sw_mdns_servant_client_node	node)
{
	if (node->m_verp == NULL)
	{
		self->m_clients = node->m_txen;

		if (node->m_txen != NULL)
		{
			node->m_txen->m_verp = NULL;
		}
	}
	else if (node->m_txen == NULL)
	{
		node->m_verp->m_txen = NULL;
	}
	else
	{
		node->m_verp->m_txen = node->m_txen;
		node->m_txen->m_verp = node->m_verp;
	}

	sw_free(node->m_name);
	sw_free(node->m_reg_type);
	sw_free(node->m_domain);
	sw_free(node->m_attrs);
	sw_free(node);

	
	return SW_OKAY;
}


static sw_result
sw_mdns_servant_convert_status_to_result(
				DNSStatus status)
{
	sw_result result;
	
	switch (status)
	{
		case kDNSNoErr:
		{
			result = SW_OKAY;
		}
		break;
		
		case kDNSNoSuchNameErr:
		{
			result = SW_DISCOVERY_E_NO_SUCH_NAME;
		}
		break;
		
		case kDNSNoMemoryErr:
		{
			result = SW_DISCOVERY_E_NO_MEM;
		}
		break;
		
		case kDNSBadParamErr:
		{
			result = SW_DISCOVERY_E_BAD_PARAM;
		}
		break;
		
		case kDNSBadReferenceErr:
		{
			result = SW_DISCOVERY_E_BAD_REFERENCE;
		}
		break;
		
		case kDNSBadStateErr:
		{
			result = SW_DISCOVERY_E_BAD_STATE;
		}
		break;
		
		case kDNSBadFlagsErr:
		{
			result = SW_DISCOVERY_E_BAD_FLAGS;
		}
		break;
		
		case kDNSUnsupportedErr:
		{
			result = SW_DISCOVERY_E_NOT_SUPPORTED;
		}
		break;
		
		case kDNSNotInitializedErr:
		{
			result = SW_DISCOVERY_E_NOT_INITIALIZED;
		}
		break;
		
		case kDNSNoCacheErr:
		{
			result = SW_DISCOVERY_E_NO_CACHE;
		}
		break;
		
		case kDNSAlreadyRegisteredErr:
		{
			result = SW_DISCOVERY_E_ALREADY_REGISTERED;
		}
		break;
		
		case kDNSNameConflictErr:
		{
			result = SW_DISCOVERY_E_NAME_CONFLICT;
		}
		break;
		
		case kDNSInvalidErr:
		{
			result = SW_DISCOVERY_E_INVALID;
		}
		break;
		
		default:
		{
			result = SW_DISCOVERY_E_UNKNOWN;
		}
	}
	
	return result;
}


static sw_result
sw_mdns_servant_getline(
					sw_mdns_servant	self,
					FILE				*	f,
					sw_string			line,
					sw_ulong			*	len)
{
	char ch;

	*len = 0;

	while (1)
	{
		ch = fgetc(f);

		if (ch == EOF)
		{
			if (*len != 0)
			{
				return SW_OKAY;
			}
			else
			{
				return SW_E_EOF;
			}
		}
		else
		{
			line[*len] = ch;
			(*len)++;

			if (ch == '\n')
			{
				return SW_OKAY;
			}
		}
	}
}


static sw_result
sw_mdns_servant_next_token(
					sw_mdns_servant	self,
					sw_string			line,
					sw_ulong			*	index,
					sw_ulong				len,
					sw_string			token)
{
	sw_ulong tokenIndex = 0;

	while ((isspace(line[*index])) && ((*index) < len))
	{
		(*index)++;
	}

	if ((*index) == len)
	{
		return SW_E_FAIL;
	}

	if (line[*index] == '"')
	{
		(*index)++;

		while ((line[*index] != '"') && ((*index) < len))
		{
			token[tokenIndex] = line[*index];
			tokenIndex++;
			(*index)++;
		}

		(*index)++;
	}
	else
	{
		while ((!isspace(line[*index])) && ((*index) < len))
		{
			token[tokenIndex] = line[*index];
			tokenIndex++;
			(*index)++;
		}
	}

	token[tokenIndex] = '\0';

	return SW_OKAY;
}


static sw_result
sw_mdns_servant_parse_conf_line(
					sw_mdns_servant	self,
					sw_string			line,
					sw_ulong				len,
					sw_string			name,
					sw_string			type,
					sw_string			domain,
					sw_ushort		*	port,
					sw_text_record	*	text_record)
{
	sw_ulong index = 0;
	sw_char	temp[256];

	if (sw_mdns_servant_next_token(self, line, &index, len, name) != SW_OKAY)
	{
		return SW_E_FAIL;
	}

	if (sw_mdns_servant_next_token(self, line, &index, len, type) != SW_OKAY)
	{
		return SW_E_FAIL;
	}

	if (sw_mdns_servant_next_token(self, line, &index, len, domain) != SW_OKAY)
	{
		return SW_E_FAIL;
	}
	
	if (sw_mdns_servant_next_token(self, line, &index, len, temp) != SW_OKAY)
	{
		return SW_E_FAIL;
	}

	(*port) = (sw_ushort) atoi(temp);

	if (sw_text_record_init(text_record) != SW_OKAY)
	{
		return SW_E_FAIL;
	}

	while (sw_mdns_servant_next_token(self, line, &index, len, temp) == SW_OKAY)
	{
		sw_text_record_add_string(*text_record, temp);
	}

	return SW_OKAY;
}
