/*
 * Copyright 2003, 2004 Porchdog Software. All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification,
 *	are permitted provided that the following conditions are met:
 *
 *		1. Redistributions of source code must retain the above copyright notice,
 *		   this list of conditions and the following disclaimer.   
 *		2. Redistributions in binary form must reproduce the above copyright notice,
 *		   this list of conditions and the following disclaimer in the documentation
 *		   and/or other materials provided with the distribution.
 *
 *	THIS SOFTWARE IS PROVIDED BY PORCHDOG SOFTWARE ``AS IS'' AND ANY
 *	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *	IN NO EVENT SHALL THE HOWL PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 *	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 *	OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	The views and conclusions contained in the software and documentation are those
 *	of the authors and should not be interpreted as representing official policies,
 *	either expressed or implied, of Porchdog Software.
 */

#ifndef _notosx_mdns_h
#define _notosx_mdns_h


#include <mdns.h>
#include <corby/orb.h>
#include <corby/object.h>
#include <salt/assert.h>
#include <discovery_i.h>
#include <NotOSX/mDNSClientAPI.h>
#include <NotOSX/DNSServices.h>


#if defined(__cplusplus)
extern "C"
{
#endif


#define MDNS_URL			"iiop://127.0.0.1:5353/dns-sd"
#define SERVICE_OID		"DNS-SD"
#define PROTOCOL_TAG		"iiop"


struct														_sw_mdns_stub_pending_op;
typedef struct _sw_mdns_stub_pending_op		*	sw_mdns_stub_pending_op;
struct														_sw_mdns_servant_client_node;
typedef struct _sw_mdns_servant_client_node	*	sw_mdns_servant_client_node;


struct _sw_mdns_stub
{
	sw_discovery				m_discovery;
	sw_salt						m_salt;
	sw_corby_orb				m_orb;
	sw_corby_object			m_self;
	sw_bool						m_bound;
	sw_corby_object			m_service;
	sw_result					m_check_version_result;
	sw_mdns_stub_pending_op	m_pending_ops;
};


struct _sw_mdns_stub_pending_op
{
	sw_opaque									m_handler;
	sw_discovery_publish_domain_reply	m_publish_domain_reply;
	sw_discovery_publish_host_reply		m_publish_host_reply;
	sw_discovery_publish_reply				m_publish_reply;
	sw_discovery_browse_reply				m_browse_reply;
	sw_discovery_resolve_reply				m_resolve_reply;
	sw_opaque									m_extra;
	sw_ulong										m_id;
	struct _sw_mdns_stub_pending_op	*	m_verp;
	struct _sw_mdns_stub_pending_op	*	m_txen;
};


struct _sw_mdns_servant
{
	sw_salt								m_salt;
	sw_corby_orb						m_orb;
	sw_ushort							m_port;
	sw_string						*	m_filters;
	sw_ulong								m_num_filters;
	sw_mdns_servant_client_node	m_clients;
};


typedef sw_result
(*sw_mdns_servant_cleanup_func)(
						sw_mdns_servant	servant,
						sw_opaque			id);


struct _sw_mdns_servant_client_node
{
	sw_mdns_servant								m_self;
	sw_string										m_host_name;
	sw_ipv4_address								m_host_address;
	sw_string										m_name;
	sw_string										m_reg_type;
	sw_string										m_domain;
	sw_port											m_port;
	sw_string										m_attrs;
	DNSDomainRegistrationRef					m_domain_registration_ref;
	DNSHostRegistrationRef						m_host_registration_ref;
	DNSRegistrationRef							m_registration_ref;
	DNSBrowserRef									m_browser_ref;
	DNSResolverRef									m_resolver_ref;
	sw_opaque										m_handler;
	sw_discovery_publish_domain_reply		m_publish_domain_reply;
	sw_discovery_publish_host_reply			m_publish_host_reply;
	sw_discovery_publish_reply					m_publish_reply;
	sw_discovery_browse_reply					m_browse_reply;
	sw_discovery_resolve_reply					m_resolve_reply;
	sw_opaque										m_extra;
	sw_ulong											m_id;
	struct _sw_mdns_servant_client_node	*	m_verp;
	struct _sw_mdns_servant_client_node	*	m_txen;

	/*
	 * this stuff is for cleaning up on client
	 * termination
	 */
	sw_corby_channel								m_channel;

	sw_result										(*m_cleanup_internal)(
															sw_mdns_servant						servant,
															struct _sw_mdns_servant_client_node	*	node);

	sw_mdns_servant_cleanup_func				m_cleanup;
};


sw_result
sw_mdns_servant_cleanup_client(
						sw_mdns_servant		self,
						sw_corby_channel		channel);


sw_result
sw_mdns_servant_publish_domain(
						sw_mdns_servant						self,
						sw_const_string						name,
						const sw_corby_object				client,
						sw_corby_channel						channel,
						sw_discovery_publish_domain_id	id);


sw_result
sw_mdns_servant_stop_publish_domain(
						sw_mdns_servant						self,
						sw_corby_channel						channel,
						sw_discovery_publish_domain_id	id);


sw_result
sw_mdns_servant_publish_host(
						sw_mdns_servant						self,
						sw_const_string						name,
						sw_const_string						domain,
						sw_ipv4_address						address,
						const sw_corby_object				client,
						sw_corby_channel						channel,
						sw_discovery_publish_domain_id	id);



sw_result
sw_mdns_servant_stop_publish_host(
						sw_mdns_servant					self,
						sw_corby_channel					channel,
						sw_discovery_publish_host_id	id);


sw_result
sw_mdns_servant_publish(
						sw_mdns_servant				self,
						sw_const_string				name,
						sw_const_string				reg_type,
						sw_const_string				domain,
						sw_const_string				host,
						sw_port							port,
						sw_octets						text_record,
						sw_ulong							text_record_len,
						const sw_corby_object		client,
						sw_corby_channel				channel,
						sw_discovery_publish_id		id);


sw_result
sw_mdns_servant_publish_update(
						sw_mdns_servant				self,
						sw_corby_channel				channel,
						sw_discovery_publish_id		id,
						sw_octets						text_record,
						sw_ulong							text_record_len);


sw_result
sw_mdns_servant_stop_publish(
						sw_mdns_servant				self,
						sw_corby_channel				channel,
						sw_discovery_publish_id		id);


sw_result
sw_mdns_servant_browse_domains(
						sw_mdns_servant				self,
						const sw_corby_object		client,
						sw_corby_channel				channel,
						sw_discovery_browse_id		id);
						
						
sw_result
sw_mdns_servant_stop_browse_domains(
						sw_mdns_servant				self,
						sw_corby_channel				channel,
						sw_discovery_browse_id		id);


sw_result
sw_mdns_servant_browse_services(
						sw_mdns_servant				self,
						sw_const_string				type,
						sw_const_string				domain,
						const sw_corby_object		client,
						sw_corby_channel				channel,
						sw_discovery_browse_id		id);
						
						
sw_result
sw_mdns_servant_stop_browse_services(
						sw_mdns_servant				self,
						sw_corby_channel				channel,
						sw_discovery_browse_id		id);


sw_result
sw_mdns_servant_resolve(
						sw_mdns_servant				self,
						sw_const_string				name,
						sw_const_string				reg_type,
						sw_const_string				domain,
						const sw_corby_object		client,
						sw_corby_channel				channel,
						sw_discovery_resolve_id		id);


sw_result
sw_mdns_servant_stop_resolve(
						sw_mdns_servant				self,
						sw_corby_channel				channel,
						sw_discovery_resolve_id		id);


sw_ushort
sw_mdns_servant_mDNSResponder_port(
						sw_mdns_servant				self);


#if defined(__cplusplus)
}
#endif


#endif
