/*
 * Copyright 2003, 2004 Porchdog Software. All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification,
 *	are permitted provided that the following conditions are met:
 *
 *		1. Redistributions of source code must retain the above copyright notice,
 *		   this list of conditions and the following disclaimer.   
 *		2. Redistributions in binary form must reproduce the above copyright notice,
 *		   this list of conditions and the following disclaimer in the documentation
 *		   and/or other materials provided with the distribution.
 *
 *	THIS SOFTWARE IS PROVIDED BY PORCHDOG SOFTWARE ``AS IS'' AND ANY
 *	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *	IN NO EVENT SHALL THE HOWL PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 *	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 *	OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	The views and conclusions contained in the software and documentation are those
 *	of the authors and should not be interpreted as representing official policies,
 *	either expressed or implied, of Porchdog Software.
 */

#include "buffer_i.h"
#include "object_i.h"
#include <salt/vstdlib.h>
#include <salt/vstring.h>
#include <salt/verrno.h>
#include <salt/assert.h>
#include <salt/log.h>

#define SW_CORBY_BUFFER_PUT_OCTET(buffer, val)			\
if (buffer->m_eptr < buffer->m_end)							\
{																		\
	*buffer->m_eptr = val;										\
	buffer->m_eptr++;												\
}																		\
else																	\
{																		\
	sw_result result;												\
	result = sw_corby_buffer_overflow(buffer, val);		\
	if (result != SW_OKAY)										\
	{																	\
		return result;												\
	}																	\
}

#define SW_CORBY_BUFFER_GET_OCTET(buffer, val)			\
if (buffer->m_bptr < buffer->m_eptr)						\
{																		\
	*val = *buffer->m_bptr;										\
	buffer->m_bptr++;												\
}																		\
else																	\
{																		\
	sw_result result;												\
	result = sw_corby_buffer_underflow(buffer, val);	\
	if (result != SW_OKAY)										\
	{																	\
		return result;												\
	}																	\
}

/*
   private stuff
*/
static sw_result
sw_corby_buffer_overflow(
                  sw_corby_buffer	buffer,
                  sw_octet				val);

static sw_result
sw_corby_buffer_underflow(
                  sw_corby_buffer	buffer,
                  sw_octet			*	val);
                  

static sw_cdr_alignment
g_cdr_align_4[4] =
{
	{ 0, 0 },
	{ 1, 3 },
	{ 2, 2 },
	{ 3, 1 }
};


static sw_cdr_alignment
g_cdr_align_8[8] =
{
	{ 0, 0 },
	{ 1, 7 },
	{ 2, 6 },
	{ 3, 5 },
	{ 4, 4 },
	{ 5, 3 },
	{ 6, 2 },
	{ 7, 1 }
};


#define SW_PAD_2(val) (((sw_ulong) (val)) % 2)
#define SW_PAD_4(val) (g_cdr_align_4[(((sw_ulong) (val)) % 4)].m_pad)
#define SW_PAD_8(val) (g_cdr_align_8[(((sw_ulong) (val)) % 8)].m_pad)


const int PANDORA_RANDOM_OID_LEN = 8;


sw_result HOWL_DLL
sw_corby_buffer_init(
					sw_corby_buffer	*	buffer)
{
   SW_TRY(sw_malloc(*buffer, sw_corby_buffer, sizeof(struct _sw_corby_buffer)));

	(*buffer)->m_base					=	NULL;
	(*buffer)->m_end					=	NULL;
	(*buffer)->m_bptr					=	NULL;
	(*buffer)->m_eptr					=	NULL;
	(*buffer)->m_delegate			=	NULL;
	(*buffer)->m_overflow_func		=	NULL;
	(*buffer)->m_underflow_func	=	NULL;
	(*buffer)->m_delegate_extra	=	NULL;
	(*buffer)->m_observer			=	NULL;
	(*buffer)->m_written_func		=	NULL;
	(*buffer)->m_observer_extra	=	NULL;
   
	return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_init_with_size(
               sw_corby_buffer	*	buffer,
               sw_ulong				size)
{
   SW_TRY(sw_malloc(*buffer, sw_corby_buffer, sizeof(struct _sw_corby_buffer)));
   
   if (sw_malloc((*buffer)->m_base, sw_octet*, size) != SW_OKAY)
   {
      sw_free(*buffer);
      return SW_E_MEM;
   }
      
   (*buffer)->m_end			=	(*buffer)->m_base + size;
   (*buffer)->m_prev			=	NULL;
   (*buffer)->m_next			=	NULL;
   (*buffer)->m_bptr			=	(*buffer)->m_base;
	(*buffer)->m_eptr			=	(*buffer)->m_base;
	(*buffer)->m_delegate			=	NULL;
	(*buffer)->m_overflow_func		=	NULL;
	(*buffer)->m_underflow_func	=	NULL;
	(*buffer)->m_delegate_extra	=	NULL;
	(*buffer)->m_observer			=	NULL;
	(*buffer)->m_written_func		=	NULL;
	(*buffer)->m_observer_extra	=	NULL;

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_init_with_delegate(
					sw_corby_buffer					*	buffer,
					sw_corby_buffer_delegate			delegate,
					sw_corby_buffer_overflow_func		overflow,
					sw_corby_buffer_underflow_func	underflow,
					sw_opaque								extra)
{
	SW_TRY(sw_corby_buffer_init(buffer));
		
	(*buffer)->m_delegate			=	delegate;
	(*buffer)->m_overflow_func		=	overflow;
	(*buffer)->m_underflow_func	=	underflow;
	(*buffer)->m_delegate_extra	=	extra;
	(*buffer)->m_observer			=	NULL;
	(*buffer)->m_written_func		=	NULL;
	(*buffer)->m_observer_extra	=	NULL;

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_init_with_size_and_delegate(
					sw_corby_buffer					*	buffer,
					sw_ulong									size,
					sw_corby_buffer_delegate			delegate,
					sw_corby_buffer_overflow_func		overflow,
					sw_corby_buffer_underflow_func	underflow,
					sw_opaque								extra)
{
	SW_TRY(sw_corby_buffer_init_with_size(buffer, size));

	(*buffer)->m_delegate			=	delegate;
	(*buffer)->m_overflow_func		=	overflow;
	(*buffer)->m_underflow_func	=	underflow;
	(*buffer)->m_delegate_extra	=	extra;
	(*buffer)->m_observer			=	NULL;
	(*buffer)->m_written_func		=	NULL;
	(*buffer)->m_observer_extra	=	NULL;

	return SW_OKAY;
}

sw_result HOWL_DLL
sw_corby_buffer_fina(
               sw_corby_buffer	self)
{
	if (self->m_base)
	{
   	sw_free(self->m_base);
	}

   sw_free(self);

   return SW_OKAY;
}


void HOWL_DLL
sw_corby_buffer_reset(
				sw_corby_buffer	self)
{
	self->m_bptr = self->m_eptr = self->m_base;
}


sw_result HOWL_DLL
sw_corby_buffer_set_octets(
				sw_corby_buffer	self,
				sw_octets			octets,
				sw_ulong				size)
{
	self->m_base 	=	octets;
	self->m_bptr 	= 	self->m_base;
	self->m_eptr 	= 	self->m_base + size;
	self->m_end		=	self->m_base + size;
	
	return SW_OKAY;
}
				

sw_octets HOWL_DLL
sw_corby_buffer_octets(
				sw_corby_buffer	self)
{
	return self->m_base;
}
				
				
sw_ulong HOWL_DLL
sw_corby_buffer_bytes_used(
				sw_corby_buffer	self)
{
	return (sw_ulong) (self->m_eptr - self->m_bptr);
}

				
sw_ulong HOWL_DLL
sw_corby_buffer_size(
				sw_corby_buffer	self)
{
	return (sw_ulong) (self->m_end - self->m_base);
}


sw_result HOWL_DLL
sw_corby_buffer_put_octet(
                  sw_corby_buffer	buffer,
                  sw_octet				val)
{
	SW_CORBY_BUFFER_PUT_OCTET(buffer, val);
	return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_put_char(
                  sw_corby_buffer	buffer,
                  sw_char				val)
{
	SW_CORBY_BUFFER_PUT_OCTET(buffer, val);
	return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_put_ushort(
               sw_corby_buffer	buffer,
               sw_ushort			val)
{
	sw_octets octets = (sw_octets) &val;

	SW_CORBY_BUFFER_PUT_OCTET(buffer, octets[0]);
	SW_CORBY_BUFFER_PUT_OCTET(buffer, octets[1]);

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_put_short(
               sw_corby_buffer	buffer,
               sw_short			val)
{
	sw_octets octets = (sw_octets) &val;

	SW_CORBY_BUFFER_PUT_OCTET(buffer, octets[0]);
	SW_CORBY_BUFFER_PUT_OCTET(buffer, octets[1]);

   return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_put_ulong(
               sw_corby_buffer	buffer,
               sw_ulong			val)
{
	sw_octets octets = (sw_octets) &val;

	SW_CORBY_BUFFER_PUT_OCTET(buffer, octets[0]);
	SW_CORBY_BUFFER_PUT_OCTET(buffer, octets[1]);
	SW_CORBY_BUFFER_PUT_OCTET(buffer, octets[2]);
	SW_CORBY_BUFFER_PUT_OCTET(buffer, octets[3]);
	
   return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_put_long(
               sw_corby_buffer	buffer,
               sw_long			val)
{
	sw_octets octets = (sw_octets) &val;

	SW_CORBY_BUFFER_PUT_OCTET(buffer, octets[0]);
	SW_CORBY_BUFFER_PUT_OCTET(buffer, octets[1]);
	SW_CORBY_BUFFER_PUT_OCTET(buffer, octets[2]);
	SW_CORBY_BUFFER_PUT_OCTET(buffer, octets[3]);
	
   return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_put_buf(
						sw_corby_buffer	buffer,
						sw_const_octets	val,
						sw_ulong				len)
{
	while (len)
	{
		sw_ulong size_left = buffer->m_end - buffer->m_eptr;

		if (size_left > 0)
		{
			sw_ulong min_bytes = (size_left < len) ? size_left : len;

			sw_memcpy(buffer->m_eptr, val, min_bytes);
			buffer->m_eptr += min_bytes;
			val				+=	min_bytes;
			len				-= min_bytes;
		}
		else
		{
			SW_TRY(sw_corby_buffer_overflow(buffer, *val));
			val++;
			len--;
		}
	}

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_put_seq(
               sw_corby_buffer	buffer,
					sw_const_octets	val,
               sw_ulong				len)
{
	SW_TRY(sw_corby_buffer_put_ulong(buffer, len));
	SW_TRY(sw_corby_buffer_put_buf(buffer, val, len));
	
   return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_put_cstring(
               sw_corby_buffer	buffer,
               sw_const_string	val)
{
   sw_ulong len = (val) ? sw_strlen(val) + 1 : 0;

	return sw_corby_buffer_put_seq(buffer, val, len);
}


sw_result
sw_corby_buffer_put_profile(
					sw_corby_buffer			self,
					const sw_corby_profile	profile)
{
	sw_result	result;
	sw_ulong		offset;
	sw_ulong		encap_size;
	sw_char		host[16];
	
	/*
	 * first thing is write tag
	 */
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(self, profile->m_tag));
	
	/*
	 * stash the offset between our index ptr and the base of the buffer.
	 * we'll use this to figure out where to write the encapsulation
	 */
	offset = (sw_ulong) (self->m_eptr - self->m_base);
	self->m_eptr += sizeof(sw_ulong);
	
	/*
	 * write the endian
	 */
	SW_TRY_GOTO(sw_corby_buffer_put_octet(self, SW_ENDIAN));
	
	/*
	 * major
	 */
	SW_TRY_GOTO(sw_corby_buffer_put_char(self, profile->m_major));
	
	/*
	 * minor
	 */
	SW_TRY_GOTO(sw_corby_buffer_put_char(self, profile->m_minor));
	
	/*
	 * host
	 */
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(self, sw_ipv4_address_name(profile->m_address, host, 16)));
	
	/*
	 * port
	 */
	SW_TRY_GOTO(sw_corby_buffer_put_ushort(self, profile->m_port));
	
	/*
	 * length of oid
	 */
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(self, profile->m_oid_len));
	
	/*
	 * oid
	 */
	SW_TRY_GOTO(sw_corby_buffer_put_buf(self, profile->m_oid, profile->m_oid_len));
	
	/*
	 * now write the length of this encap
	 * hold on..this gets a little hairy
	 */
	encap_size = (sw_ulong) (self->m_eptr - self->m_base) - offset - sizeof(sw_ulong);
	sw_memcpy((self->m_base + offset), &encap_size, sizeof(sw_ulong));

	/*
	 * if we got here, we're good
	 */
	result = SW_OKAY;
	
exit:

	if (result != SW_OKAY)
	{
	}

	return result;
}
					

sw_result
sw_corby_buffer_put_ior(
					sw_corby_buffer		self,
					const sw_corby_ior	ior)
{
	sw_corby_profile	profile;
	sw_result			result;
	
	/*
	 * repository id
	 */
	SW_TRY_GOTO(sw_corby_buffer_put_cstring(self, ior->m_repository_id));
	
	/*
	 * number of profiles
	 */
	SW_TRY_GOTO(sw_corby_buffer_put_ulong(self, ior->m_num_profiles));
	
	/*
	 * the profiles
	 */
	for (profile = ior->m_profiles; profile; profile = profile->m_txen)
	{
		SW_TRY_GOTO(sw_corby_buffer_put_profile(self, profile));
	}
	
	result = SW_OKAY;
	
exit:

	return result;
}


sw_result HOWL_DLL
sw_corby_buffer_put_object(
					sw_corby_buffer						self,
					const struct _sw_corby_object	*	object)
{
	SW_TRY(sw_corby_buffer_put_ior(self, object->m_iors));
	return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_put_pad(
				sw_corby_buffer						self,
				sw_corby_buffer_pad					pad)
{
	sw_ulong bytes_used = (sw_ulong) (self->m_eptr - self->m_bptr);

	switch (pad)
	{
		case SW_CORBY_BUFFER_PAD_ALIGN_2:
		{
			self->m_eptr += (bytes_used % 2);
		}
		break;

		case SW_CORBY_BUFFER_PAD_ALIGN_4:
		{
			sw_ulong rem = (bytes_used % 4);

			self->m_eptr += (rem > 0) ? (4 - rem) : 0;
		}
		break;

		case SW_CORBY_BUFFER_PAD_ALIGN_8:
		{
			sw_ulong rem = (bytes_used % 8);

			self->m_eptr += (rem > 0) ? (8 - rem) : 0;
		}
		break;

		case SW_CORBY_BUFFER_PAD_ALIGN_16:
		{
			sw_ulong rem = (bytes_used % 16);

			self->m_eptr += (rem > 0) ? (16 - rem) : 0;
		}
		break;

		case SW_CORBY_BUFFER_PAD_ALIGN_32:
		{
			sw_ulong rem = (bytes_used % 32);

			self->m_eptr += (rem > 0) ? (32 - rem) : 0;
		}
		break;
	}

	sw_assert(self->m_eptr <= self->m_end);

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_get_octet(
                  sw_corby_buffer	buffer,
                  sw_octet			*	val)
{
	SW_CORBY_BUFFER_GET_OCTET(buffer, val);
	return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_get_char(
               sw_corby_buffer	buffer,
               sw_char			*	val)
{
	SW_CORBY_BUFFER_GET_OCTET(buffer, val);
	return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_get_ushort(
               sw_corby_buffer	buffer,
               sw_ushort		*	val,
               sw_octet				endian)
{
	sw_octets octets = (sw_octets) val;

	if (endian == SW_ENDIAN)
	{
		SW_CORBY_BUFFER_GET_OCTET(buffer, octets);
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 1));
	}
	else
	{
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 1));
		SW_CORBY_BUFFER_GET_OCTET(buffer, octets);
	}

   return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_get_short(
               sw_corby_buffer		buffer,
               sw_short			*	val,
               sw_octet				endian)
{
	sw_octets octets = (sw_octets) val;

	if (endian == SW_ENDIAN)
	{
		SW_CORBY_BUFFER_GET_OCTET(buffer, octets);
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 1));
	}
	else
	{
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 1));
		SW_CORBY_BUFFER_GET_OCTET(buffer, octets);
	}
	
   return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_get_ulong(
               sw_corby_buffer		buffer,
               sw_ulong			*	val,
               sw_octet				endian)
{
	sw_octets octets = (sw_octets) val;

	if (endian == SW_ENDIAN)
	{
		SW_CORBY_BUFFER_GET_OCTET(buffer, octets);
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 1));
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 2));
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 3));
	}
	else
	{
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 3));
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 2));
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 1));
		SW_CORBY_BUFFER_GET_OCTET(buffer, octets);
	}
	
   return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_get_long(
               sw_corby_buffer		buffer,
               sw_long			*	val,
               sw_octet				endian)
{
	sw_octets octets = (sw_octets) val;

	if (endian == SW_ENDIAN)
	{
		SW_CORBY_BUFFER_GET_OCTET(buffer, octets);
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 1));
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 2));
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 3));
	}
	else
	{
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 3));
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 2));
		SW_CORBY_BUFFER_GET_OCTET(buffer, (octets + 1));
		SW_CORBY_BUFFER_GET_OCTET(buffer, octets);
	}
	
   return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_get_buf(
				sw_corby_buffer	buffer,
				sw_octets			buf,
				sw_ulong				len)
{
	while (len)
	{
		sw_ulong bytes_in_buf = (buffer->m_eptr - buffer->m_bptr);

		if (bytes_in_buf > 0)
		{
			sw_ulong min_bytes = (bytes_in_buf < len) ? bytes_in_buf : len;

			sw_memcpy(buf, buffer->m_bptr, min_bytes);
			buffer->m_bptr += min_bytes;
			buf				+=	min_bytes;
			len				-= min_bytes;
		}
		else
		{
			SW_TRY(sw_corby_buffer_underflow(buffer, buf));
			buf++;
			len--;
		}
	}

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_allocate_and_get_seq(
					sw_corby_buffer	self,
					sw_octets		*	val,
					sw_ulong			*	len,
					sw_octet				endian)
{
	SW_TRY(sw_corby_buffer_get_ulong(self, len, endian));
	SW_TRY(sw_malloc((*val), sw_octets, *len));
	return sw_corby_buffer_get_buf(self, *val, *len);
}


sw_result HOWL_DLL
sw_corby_buffer_get_zerocopy_seq(
					sw_corby_buffer	self,
					sw_octets		*	val,
					sw_ulong			*	len,
					sw_octet				endian)
{
	sw_ulong bytes_in_buf;

	SW_TRY(sw_corby_buffer_get_ulong(self, len, endian));

	bytes_in_buf = (self->m_eptr - self->m_bptr);

	if (*len > bytes_in_buf)
	{
		return SW_E_FAIL;
	}

	if (*len > 0)
	{
		*val = self->m_bptr;
		self->m_bptr += *len;
	}
	else
	{
		*val = NULL;
	}

	return SW_OKAY;
}
                                    

sw_result HOWL_DLL
sw_corby_buffer_get_seq(
					sw_corby_buffer	buffer,
					sw_octets			val,
					sw_ulong			*	len,
					sw_octet				endian)
{
	sw_ulong max = *len;

	sw_corby_buffer_get_ulong(buffer, len, endian);

	if (max < *len)
	{
		return SW_E_FAIL;
	}

	return sw_corby_buffer_get_buf(buffer, val, *len);
}


sw_result HOWL_DLL
sw_corby_buffer_allocate_and_get_cstring(
					sw_corby_buffer	self,
					sw_string		*	val,
					sw_ulong			*	len,
					sw_octet				endian)
{
	SW_TRY(sw_corby_buffer_get_ulong(self, len, endian));
	SW_TRY(sw_malloc((*val), sw_string, *len));
	return sw_corby_buffer_get_buf(self, (sw_octets) *val, *len);
}
				

sw_result HOWL_DLL
sw_corby_buffer_get_zerocopy_cstring(
					sw_corby_buffer	self,
					sw_string		*	val,
					sw_ulong			*	len,
					sw_octet				endian)
{
	sw_ulong bytes_in_buf;

	SW_TRY(sw_corby_buffer_get_ulong(self, len, endian));

	bytes_in_buf = (self->m_eptr - self->m_bptr);

	if (*len > bytes_in_buf)
	{
		return SW_E_FAIL;
	}

	if (*len > 0)
	{
		*val = self->m_bptr;
		self->m_bptr += *len;
	}
	else
	{
		*val = NULL;
	}

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_corby_buffer_get_cstring(
               sw_corby_buffer	buffer,
               sw_string			val,
               sw_ulong			*	len,
               sw_octet				endian)
{
	sw_ulong max = *len;

	SW_TRY(sw_corby_buffer_get_ulong(buffer, len, endian));

	if (max < *len)
	{
		return SW_E_FAIL;
	}

	return sw_corby_buffer_get_buf(buffer, val, *len);
}


sw_result
sw_corby_buffer_get_profile(
					sw_corby_buffer		self,
					sw_corby_profile	*	profile,
					sw_octet					endian)
{
	sw_result	result;
	sw_ulong 	encap_size;
	sw_octet		encap_endian;
	sw_char		host[16];
	sw_ulong		host_len;
	
	/*
	 * allocate storage
	 */
	SW_TRY_GOTO(sw_corby_profile_init(profile));
	
	/*
	 * tag
	 */
	SW_TRY_GOTO(sw_corby_buffer_get_ulong(self, &(*profile)->m_tag, endian));
	
	/*
	 * size
	 */
	SW_TRY_GOTO(sw_corby_buffer_get_ulong(self, &encap_size, endian));
	
	/*
	 * endian
	 */
	SW_TRY_GOTO(sw_corby_buffer_get_octet(self, &encap_endian));
	
	/*
	 * major
	 */
	SW_TRY_GOTO(sw_corby_buffer_get_char(self, &(*profile)->m_major));
	
	/*
	 * minor
	 */
	SW_TRY_GOTO(sw_corby_buffer_get_char(self, &(*profile)->m_minor));
	
	/*
	 * host
	 */
	host_len = 16;
	SW_TRY_GOTO(sw_corby_buffer_get_cstring(self, host, &host_len, encap_endian));
	SW_TRY_GOTO(sw_ipv4_address_init_from_name(&(*profile)->m_address, host));
	
	/*
	 * port
	 */
	SW_TRY_GOTO(sw_corby_buffer_get_ushort(self, &(*profile)->m_port, endian));
	
	/*
	 * oid_len and oid
	 */
	SW_TRY_GOTO(sw_corby_buffer_allocate_and_get_seq(self, (sw_octets*) &(*profile)->m_oid, &(*profile)->m_oid_len, encap_endian));

	/*
	 * if we got here, then we're okay
	 */
	result = SW_OKAY;

exit:

	if (result != SW_OKAY)
	{
	}
	
	return result;
}
					
					
sw_result
sw_corby_buffer_get_ior(
					sw_corby_buffer		self,
					sw_corby_ior		*	ior,
					sw_octet					endian)
{
	sw_result			result;
	sw_corby_profile	profile;
	sw_corby_profile	last;
	sw_ulong				len;
	sw_ulong				i;
	
	/*
	 * allocate
	 */
	SW_TRY_GOTO(sw_corby_ior_init(ior));
	
	/*
	 * repository id
	 */
	SW_TRY_GOTO(sw_corby_buffer_allocate_and_get_cstring(self, &(*ior)->m_repository_id, &len, endian));
	
	/*
	 * number of profiles
	 */
	SW_TRY_GOTO(sw_corby_buffer_get_ulong(self, &(*ior)->m_num_profiles, endian));
	
	/*
	 * profiles...we want to store in the order they were read
	 */
	for (i = 0, last = NULL; i < (*ior)->m_num_profiles; i++)
	{		
		SW_TRY_GOTO(sw_corby_buffer_get_profile(self, &profile, endian));
		
		if (last == NULL)
		{
			(*ior)->m_profiles = profile;
		}
		else
		{
			last->m_txen	=	profile;
		}

		last = profile;
	}
	
	/*
	 * if we got here, we're good
	 */
	result = SW_OKAY;
	
exit:

	if (result != SW_OKAY)
	{
	}
	
	return result;
}


sw_result HOWL_DLL
sw_corby_buffer_get_object(
					sw_corby_buffer		self,
					sw_corby_object	*	object,
					sw_octet					endian)
{
	sw_result result;
	
	/*
	 * allocate
	 */
	SW_TRY_GOTO(sw_corby_object_init(object));
	
	/*
	 * ior
	 */
	SW_TRY_GOTO(sw_corby_buffer_get_ior(self, &(*object)->m_iors, endian));
	
	/*
	 * if we got here, we're good
	 */
	result = SW_OKAY;
	
exit:

	if (result != SW_OKAY)
	{
	}
	
	return result;
}


static sw_result
sw_corby_buffer_overflow(
						sw_corby_buffer	buffer,
						sw_octet				val)
{
	if (buffer->m_overflow_func != NULL)
	{
		SW_TRY((*buffer->m_overflow_func)(buffer->m_delegate, buffer, val, &buffer->m_base, &buffer->m_bptr, &buffer->m_eptr, &buffer->m_end, buffer->m_delegate_extra));
	}
	else
	{
 		sw_ulong newlen;
 		sw_ulong oldlen;
   	sw_ulong boff;
   	sw_ulong eoff;
   
   	/*
      	get all the relevant info
   	*/
   	oldlen	=	buffer->m_end - buffer->m_base;
   	boff		=	buffer->m_bptr - buffer->m_base;
   	eoff		=	buffer->m_eptr	- buffer->m_base;
   	newlen	=	oldlen * 2;
  	 
   	/*
      	new realloc
   	*/
   	SW_TRY(sw_realloc(buffer->m_base, sw_octets, newlen));

		/*
		 * reset the pointers
		 */
   	buffer->m_bptr = buffer->m_base + boff;
   	buffer->m_eptr = buffer->m_base + eoff;
   	buffer->m_end  = buffer->m_base + newlen;

		/*
		 * and store the guy
		 */
		*(buffer->m_eptr) = val;
		buffer->m_eptr++;
	}
   
   return SW_OKAY;
}


static sw_result
sw_corby_buffer_underflow(
                  sw_corby_buffer	buffer,
                  sw_octet			*	val)
{
	if (buffer->m_underflow_func == NULL)
	{
   	sw_log(SW_LOG_ERROR, "sw_corby_buffer_underflow", "internal error\n");
   	return SW_E_CORBY_MARSHAL;
	}

	return (*buffer->m_underflow_func)(buffer->m_delegate, buffer, val, &buffer->m_base, &buffer->m_bptr, &buffer->m_eptr, &buffer->m_end, buffer->m_delegate_extra);
}
