/*
 * Copyright 2003, 2004 Porchdog Software. All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification,
 *	are permitted provided that the following conditions are met:
 *
 *		1. Redistributions of source code must retain the above copyright notice,
 *		   this list of conditions and the following disclaimer.   
 *		2. Redistributions in binary form must reproduce the above copyright notice,
 *		   this list of conditions and the following disclaimer in the documentation
 *		   and/or other materials provided with the distribution.
 *
 *	THIS SOFTWARE IS PROVIDED BY PORCHDOG SOFTWARE ``AS IS'' AND ANY
 *	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *	IN NO EVENT SHALL THE HOWL PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 *	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 *	OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	The views and conclusions contained in the software and documentation are those
 *	of the authors and should not be interpreted as representing official policies,
 *	either expressed or implied, of Porchdog Software.
 */

#include "discovery_i.h"
#include <salt/vstdlib.h>
#include <salt/verrno.h>
#include <salt/assert.h>


sw_result HOWL_DLL
sw_discovery_init(
						sw_discovery			*	self)
{
	return sw_discovery_init_with_flags(self, SW_DISCOVERY_USE_SHARED_SERVICE);
}


sw_result HOWL_DLL
sw_discovery_init_with_flags(
						sw_discovery			*	self,
						sw_discovery_init_flags	flags)
{
	if (sw_malloc((*self), sw_discovery, sizeof(struct _sw_discovery)) != SW_OKAY)
	{
		sw_log(SW_LOG_ERROR, "sw_discovery_init_with_flags", "out of memory\n");
		return SW_E_FAIL;
	}

	(*self)->m_servant	=	NULL;
	(*self)->m_stub		=	NULL;

	/*
	 * initialize salt
	 */
	if (sw_salt_init(&(*self)->m_salt, 0, NULL) != SW_OKAY)
	{
		return SW_E_INIT;
	}

	if (flags & SW_DISCOVERY_USE_SHARED_SERVICE)
	{
		if (sw_mdns_stub_init(&(*self)->m_stub, (*self)->m_salt, (*self), 5353) == SW_OKAY)
		{
			if ((flags & SW_DISCOVERY_SKIP_VERSION_CHECK) == 0)
			{
				if (sw_mdns_stub_check_version((*self)->m_stub) != SW_OKAY)
				{
					sw_mdns_stub_fina((*self)->m_stub);
					(*self)->m_stub = NULL;
				}
			}
		}
	}

	if (((*self)->m_stub == NULL) && (flags & SW_DISCOVERY_USE_PRIVATE_SERVICE))
	{
		if (sw_mdns_servant_new(&(*self)->m_servant, 0, NULL, 0) == SW_OKAY)
		{
			if (sw_mdns_stub_init(&(*self)->m_stub, (*self)->m_salt, (*self), sw_mdns_servant_port((*self)->m_servant)) != SW_OKAY)
			{
				sw_mdns_servant_delete((*self)->m_servant);
				(*self)->m_servant	=	NULL;
				(*self)->m_stub		=	NULL;
			}
		}
	}

	if ((*self)->m_stub == NULL)
	{
		return SW_E_FAIL;
	}

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_discovery_fina(
						sw_discovery		self)
{
	if (self->m_stub)
	{
		sw_mdns_stub_fina(self->m_stub);
	}

	if (self->m_servant)
	{
		sw_mdns_servant_delete(self->m_servant);
	}

	if (self->m_salt)
	{
		sw_salt_fina(self->m_salt);
	}

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_discovery_publish_domain(
						sw_discovery								self,
						sw_const_string							name,
						sw_discovery_publish_domain_handler	handler,
						sw_discovery_publish_domain_reply	reply,
						sw_opaque									extra,
						sw_discovery_publish_domain_id	*	id)
{
	if (self->m_stub != NULL)
	{
		return sw_mdns_stub_publish_domain(self->m_stub, name, handler, reply, extra, id);
	}
	else
	{
		return SW_E_FAIL;
	}
}


sw_result HOWL_DLL
sw_discovery_stop_publish_domain(
						sw_discovery							self,
						sw_discovery_publish_domain_id	id)
{
	if (self->m_stub)
	{
		return sw_mdns_stub_stop_publish_domain(self->m_stub, id);
	}
	else
	{
		return SW_E_FAIL;
	}
}


sw_result HOWL_DLL
sw_discovery_publish_host(
						sw_discovery								self,
						sw_const_string							name,
						sw_const_string							domain,
						sw_ipv4_address							address,
						sw_discovery_publish_host_handler	handler,
						sw_discovery_publish_host_reply		reply,
						sw_opaque									extra,
						sw_discovery_publish_host_id		*	id)
{
	if (self->m_stub)
	{
		return sw_mdns_stub_publish_host(self->m_stub, name, domain, address, handler, reply, extra, id);
	}
	else
	{
		return SW_E_FAIL;
	}
}


sw_result HOWL_DLL
sw_discovery_stop_publish_host(
						sw_discovery						self,
						sw_discovery_publish_host_id	id)
{
	if (self->m_stub)
	{
		return sw_mdns_stub_stop_publish_host(self->m_stub, id);
	}
	else
	{
		return SW_E_FAIL;
	}
}


sw_result HOWL_DLL
sw_discovery_publish(
						sw_discovery						self,
						sw_const_string					name,
						sw_const_string					type,
						sw_const_string					domain,
						sw_const_string					host,
						sw_port								port,
						sw_octets							text_record,
						sw_ulong								text_record_len,
						sw_discovery_publish_handler	handler,
						sw_discovery_publish_reply		reply,
						sw_opaque							extra,
						sw_discovery_publish_id		*	id)
{
	if (self->m_stub)
	{
		return sw_mdns_stub_publish(self->m_stub, name, type, domain, host, port, text_record, text_record_len, handler, reply, extra, id);
	}
	else
	{
		return SW_E_FAIL;
	}
}



sw_result HOWL_DLL
sw_discovery_publish_update(
						sw_discovery					self,
						sw_discovery_publish_id		id,
						sw_octets						text_record,
						sw_ulong							text_record_len)
{
	if (self->m_stub)
	{
		return sw_mdns_stub_publish_update(self->m_stub, id, text_record, text_record_len);
	}
	else
	{
		return SW_E_FAIL;
	}
}


sw_result HOWL_DLL
sw_discovery_stop_publish(
						sw_discovery					self,
						sw_discovery_publish_id		id)
{
	if (self->m_stub != NULL)
	{
		return sw_mdns_stub_stop_publish(self->m_stub, id);
	}
	else
	{
		return SW_E_FAIL;
	}
}


sw_result HOWL_DLL
sw_discovery_browse_domains(
						sw_discovery						self,
						sw_discovery_browse_handler	handler,
						sw_discovery_browse_reply		reply,
						sw_opaque							extra,
						sw_discovery_browse_id		*	id)
{
	if (self->m_stub != NULL)
	{
		return sw_mdns_stub_browse_domains(self->m_stub, handler, reply, extra, id);
	}
	else
	{
		return SW_E_FAIL;
	}
}


sw_result HOWL_DLL
sw_discovery_stop_browse_domains(
						sw_discovery						self,
						sw_discovery_browse_id			id)
{
	if (self->m_stub != NULL)
	{
		return sw_mdns_stub_stop_browse_domains(self->m_stub, id);
	}
	else
	{
		return SW_E_FAIL;
	}
}


sw_result HOWL_DLL
sw_discovery_browse(
						sw_discovery						self,
						sw_const_string					type,
						sw_const_string					domain,
						sw_discovery_browse_handler	handler,
						sw_discovery_browse_reply		reply,
						sw_opaque							extra,
						sw_discovery_browse_id		*	id)
{
	if (self->m_stub != NULL)
	{
		return sw_mdns_stub_browse_services(self->m_stub, type, domain, handler, reply, extra, id);
	}
	else
	{
		return SW_E_FAIL;
	}
}


sw_result HOWL_DLL
sw_discovery_stop_browse(
						sw_discovery						self,
						sw_discovery_browse_id			id)
{
	if (self->m_stub != NULL)
	{
		return sw_mdns_stub_stop_browse_services(self->m_stub, id);
	}
	else
	{
		return SW_E_FAIL;
	}
}


sw_result HOWL_DLL
sw_discovery_resolve(
						sw_discovery						self,
						sw_const_string					name,
						sw_const_string					type,
						sw_const_string					domain,
						sw_discovery_resolve_handler	handler,
						sw_discovery_resolve_reply		reply,
						sw_opaque							extra,
						sw_discovery_resolve_id		*	id)
{
	if (self->m_stub != NULL)
	{
		return sw_mdns_stub_resolve(self->m_stub, name, type, domain, handler, reply, extra, id);
	}
	else
	{
		return SW_E_FAIL;
	}
}


sw_result HOWL_DLL
sw_discovery_stop_resolve(
						sw_discovery					self,
						sw_discovery_resolve_id		id)
{
	if (self->m_stub != NULL)
	{
		return sw_mdns_stub_stop_resolve(self->m_stub, id);
	}
	else
	{
		return SW_E_FAIL;
	}
}
						

sw_result HOWL_DLL
sw_discovery_run(
						sw_discovery					self)
{
	if (self->m_salt != NULL)
	{
		return sw_salt_run(self->m_salt);
	}
	else
	{
		return SW_E_FAIL;
	}
}


sw_result HOWL_DLL
sw_discovery_stop_run(
						sw_discovery					self)
{
	if (self->m_salt != NULL)
	{
		return sw_salt_stop_run(self->m_salt);
	}
	else
	{
		return SW_E_FAIL;
	}
}
						

int HOWL_DLL
sw_discovery_socket(
						sw_discovery					self)
{
	if (self->m_stub != NULL)
	{
		return sw_mdns_stub_socket(self->m_stub);
	}
	else
	{
		return -1;
	}
}


sw_result HOWL_DLL
sw_discovery_read_socket(
						sw_discovery					self)
{
	if (self->m_stub != NULL)
	{
		return sw_mdns_stub_read_socket(self->m_stub);
	}
	else
	{
		return SW_E_FAIL;
	}
}


sw_result HOWL_DLL
sw_discovery_salt(
						sw_discovery		self,
						sw_salt			*	salt)
{
	*salt = self->m_salt;
	return SW_OKAY;
}
