/*
 * Copyright 2003, 2004 Porchdog Software. All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification,
 *	are permitted provided that the following conditions are met:
 *
 *		1. Redistributions of source code must retain the above copyright notice,
 *		   this list of conditions and the following disclaimer.   
 *		2. Redistributions in binary form must reproduce the above copyright notice,
 *		   this list of conditions and the following disclaimer in the documentation
 *		   and/or other materials provided with the distribution.
 *
 *	THIS SOFTWARE IS PROVIDED BY PORCHDOG SOFTWARE ``AS IS'' AND ANY
 *	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *	IN NO EVENT SHALL THE HOWL PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 *	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 *	OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	The views and conclusions contained in the software and documentation are those
 *	of the authors and should not be interpreted as representing official policies,
 *	either expressed or implied, of Porchdog Software.
 */

#include "text_record_i.h"
#include <salt/vstdlib.h>
#include <salt/vstring.h>
#include <salt/verrno.h>
#include <salt/assert.h>

static const sw_ulong MaxStringLen	=	255;

sw_result HOWL_DLL
sw_text_record_init(
			sw_text_record	*	self)
{
	SW_TRY(sw_malloc(*self, sw_text_record, sizeof(struct _sw_text_record)));
	SW_TRY(sw_corby_buffer_init_with_size(&((*self)->m_buffer), 1024));

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_text_record_fina(
			sw_text_record		self)
{
	sw_corby_buffer_fina(self->m_buffer);
	sw_free(self);

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_text_record_add_string(
				sw_text_record		self,
				sw_const_string	string)
{
	sw_ulong len;

	sw_assert(string != NULL);

	len = sw_strlen(string);

	if (len > MaxStringLen)
	{
		return SW_E_FAIL;
	}

	SW_TRY(sw_corby_buffer_put_octet(self->m_buffer, (sw_octet) len));
	SW_TRY(sw_corby_buffer_put_buf(self->m_buffer, string, len));

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_text_record_add_key_and_string_value(
				sw_text_record		self,
				sw_const_string	key,
				sw_const_string	val)
{
	sw_ulong len;

	sw_assert(key != NULL);

	len = sw_strlen(key) + 1;

	if (val)
	{
		len += sw_strlen(val);
	}

	if (len > MaxStringLen)
	{
		return SW_E_FAIL;
	}

	SW_TRY(sw_corby_buffer_put_octet(self->m_buffer, (sw_octet) len));
	SW_TRY(sw_corby_buffer_put_buf(self->m_buffer, key, sw_strlen(key)));
	SW_TRY(sw_corby_buffer_put_octet(self->m_buffer, '='));

	if (val)
	{
		SW_TRY(sw_corby_buffer_put_buf(self->m_buffer, val, sw_strlen(val)));
	}

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_text_record_add_key_and_opaque_value(
				sw_text_record		self,
				sw_const_string	key,
				sw_opaque			val,
				sw_ulong				val_len)
{
	sw_ulong len;

	sw_assert(key != NULL);

	len = sw_strlen(key) + 1 + val_len;

	if (len > MaxStringLen)
	{
		return SW_E_FAIL;
	}

	SW_TRY(sw_corby_buffer_put_octet(self->m_buffer, (sw_octet) len));
	SW_TRY(sw_corby_buffer_put_buf(self->m_buffer, key, sw_strlen(key)));
	SW_TRY(sw_corby_buffer_put_octet(self->m_buffer, '='));
	SW_TRY(sw_corby_buffer_put_buf(self->m_buffer, val, (sw_octet) val_len));

	return SW_OKAY;
}


sw_octets HOWL_DLL
sw_text_record_bytes(
				sw_text_record		self)
{
	return sw_corby_buffer_octets(self->m_buffer);
}


sw_ulong HOWL_DLL
sw_text_record_len(
				sw_text_record		self)
{
	return sw_corby_buffer_bytes_used(self->m_buffer);
}


sw_result HOWL_DLL
sw_text_record_iterator_init(
				sw_text_record_iterator				*	self,
				sw_octets									text_record,
				sw_ulong										text_record_len)
{
	SW_TRY(sw_malloc(*self, sw_text_record_iterator, sizeof(struct _sw_text_record_iterator)));
	SW_TRY(sw_corby_buffer_init(&((*self)->m_buffer)));
	SW_TRY(sw_corby_buffer_set_octets((*self)->m_buffer, text_record, text_record_len));

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_text_record_iterator_fina(
				sw_text_record_iterator					self)
{
	sw_corby_buffer_set_octets(self->m_buffer, NULL, 0);
	sw_corby_buffer_fina(self->m_buffer);
	sw_free(self);

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_text_record_iterator_next(
				sw_text_record_iterator					self,
				sw_char										key[SW_TEXT_RECORD_MAX_LEN],
				sw_octet										val[SW_TEXT_RECORD_MAX_LEN],
				sw_ulong									*	val_len)
{
	sw_bool	decodingValue;
	sw_octet len;
	int		i;

	if (sw_corby_buffer_octets(self->m_buffer) == NULL)
	{
		return SW_E_FAIL;
	}

	if (sw_corby_buffer_bytes_used(self->m_buffer) == 0)
	{
		return SW_E_FAIL;
	}

	memset(key, 0, SW_TEXT_RECORD_MAX_LEN);
	memset(val, 0, SW_TEXT_RECORD_MAX_LEN);

	decodingValue	=	SW_FALSE;
	*val_len			=	0;

	SW_TRY(sw_corby_buffer_get_octet(self->m_buffer, &len));
	sw_assert(len <= MaxStringLen);

	for (i = 0; i < len; i++)
	{
		sw_char c;

		SW_TRY(sw_corby_buffer_get_char(self->m_buffer, &c));

		if (decodingValue == SW_FALSE)
		{
			if (c != '=')
			{
				key[i] = c;
			}
			else
			{
				decodingValue	=	SW_TRUE;
			}
		}
		else
		{
			val[*val_len] = c;
			(*val_len)++;
		}
	}

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_text_record_string_iterator_init(
				sw_text_record_string_iterator	*	self,
				sw_const_string							text_record_string)
{
	SW_TRY(sw_malloc(*self, sw_text_record_string_iterator, sizeof(struct _sw_text_record_string_iterator)));

	(*self)->m_string =	text_record_string;
	(*self)->m_index	=	0;

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_text_record_string_iterator_fina(
				sw_text_record_string_iterator		self)
{
	sw_free(self);

	return SW_OKAY;
}


sw_result HOWL_DLL
sw_text_record_string_iterator_next(
				sw_text_record_string_iterator	self,
				sw_char									key[SW_TEXT_RECORD_MAX_LEN],
				sw_char									val[SW_TEXT_RECORD_MAX_LEN])
{
	sw_bool	decodingValue;
	sw_ulong keyIndex;
	sw_ulong	valIndex;

	if (self->m_string == NULL)
	{
		return SW_E_FAIL;
	}

	if (self->m_string[self->m_index] == '\0')
	{
		return SW_E_FAIL;
	}
	else if (self->m_string[self->m_index] == '\001')
	{
		self->m_index++;
	}

	memset(key, 0, SW_TEXT_RECORD_MAX_LEN);
	memset(val, 0, SW_TEXT_RECORD_MAX_LEN);
	decodingValue	=	SW_FALSE;
	keyIndex			=	0;
	valIndex			=	0;

	while ((self->m_string[self->m_index] != '\0') && (self->m_string[self->m_index] != '\001'))
	{
		char c = self->m_string[self->m_index++];

		if (decodingValue == SW_FALSE)
		{
			if (c != '=')
			{
				key[keyIndex++]	=	c;
			}
			else
			{
				decodingValue = SW_TRUE;
			}
		}
		else
		{
			val[valIndex++] = c;
		}
	}

	return SW_OKAY;
}